"""Janggu - deep learning for genomics"""

import copy
import hashlib
import logging
import os
import re
import time

import h5py
import keras
import keras.backend as K
import tensorflow as tf
import numpy as np
from keras.callbacks import CSVLogger
from keras.callbacks import LambdaCallback
from keras.models import Model
from keras.models import load_model
from keras.utils import Sequence
from keras.utils import plot_model
from progress.bar import Bar
from pybedtools import BedTool

from janggu.data import split_train_test
from janggu.data.coverage import Cover
from janggu.data.dna import Bioseq
from janggu.data.data import JangguSequence
from janggu.data.data import _data_props
from janggu.data.dna import VariantStreamer
from janggu.data.genomic_indexer import GenomicIndexer
from janggu.evaluation import get_scorer
from janggu.layers import Complement
from janggu.layers import DnaConv2D
from janggu.layers import LocalAveragePooling2D
from janggu.layers import Reverse
from janggu.utils import _get_output_root_directory
from janggu.utils import _to_list

JANGGU_LAYERS = {'Reverse': Reverse,
                 'Complement': Complement,
                 'LocalAveragePooling2D': LocalAveragePooling2D,
                 'DnaConv2D': DnaConv2D}


def _convert_data(kerasmodel, data, layer):
    """Converts different data formats to
    {name:values} dict.

    keras support different data formats, e.g. np.array,
    list(np.array) or dict(key: np.array).
    This function normalizes all dataset to the dictionary
    style usage internally. This simplifies the compatibility
    checks at various places.

    Parameters
    ----------
    kerasmodel : keras.Model
        A keras Model object.
    data : Dataset, np.array, list or dict
        Dataset.
    layer : str
        Indication as to whether data is used as input or output.
        :code:`layer='output_layers'` or 'input_layers'.
    """
    # If we deal with Dataset, we convert it to a Dictionary
    # which is directly interpretable by keras
    if hasattr(data, "name") and hasattr(data, "shape"):
        # if it is a janggu.data.Dataset we get here
        c_data = {data.name: data}
    elif not hasattr(data, "name") and hasattr(data, "shape"):
        # if data is a numpy array we get here
        c_data = {kerasmodel.get_config()[layer][0][0]: data}
    elif isinstance(data, list):
        if hasattr(data[0], "name"):
            # if data is a list(Dataset) we get here
            c_data = {datum.name: datum for datum in data}
        else:
            # if data is a list(np.array) we get here
            c_data = {name[0]: datum for name, datum in
                      zip(kerasmodel.get_config()[layer],
                          data)}
    elif isinstance(data, dict):
        # if data is a dict, it can just be passed through
        c_data = data
    else:
        raise ValueError('Unknown data type: {}'.format(type(data)))

    return c_data


class Janggu(object):
    """Janggu class

    The class :class:`Janggu` maintains a :class:`keras.models.Model`
    object, that is an instance of a neural network.
    Furthermore, to the outside, Janggu behaves similarly to
    :class:`keras.models.Model` which allows you to create,
    fit, and evaluate the model.

    Parameters
    -----------
    inputs : Input or list(Input)
        Input layer or list of Inputs as defined by keras.
        See https://keras.io/.
    outputs : Layer or list(Layer)
        Output layer or list of outputs. See https://keras.io/.
    name : str
        Name of the model.

    Examples
    --------

    Define a Janggu object similar to keras.models.Model
    using Input and Output layers.

    .. code-block:: python

      from keras.layers import Input
      from keras.layers import Dense

      from janggu import Janggu

      # Define neural network layers using keras
      in_ = Input(shape=(10,), name='ip')
      layer = Dense(3)(in_)
      output = Dense(1, activation='sigmoid', name='out')(layer)

      # Instantiate a model.
      model = Janggu(inputs=in_, outputs=output, name='test_model')
      model.summary()
    """
    timer = None
    _name = None

    def __init__(self, inputs, outputs, name=None):

        self.kerasmodel = Model(inputs, outputs, name='janggu')

        if not name:

            hasher = hashlib.md5()
            hasher.update(self.kerasmodel.to_json().encode('utf-8'))
            name = hasher.hexdigest()
            print("Generated model-id: '{}'".format(name))

        if hasattr(outputs, '__len__') > 1:
            total_output = K.sum([K.sum(o, axis=-1) for o in self.kerasmodel.output], axis=-1)
        else:
            total_output = K.sum(self.kerasmodel.output, axis=-1)

        kinp = self.kerasmodel.input
        kinp = _to_list(kinp)

        if tf.__version__[0] == '1':
            grad = K.gradients(total_output, self.kerasmodel.input)

            self._influence = K.function(kinp, grad)
        else:
            def _influence(inputs):
                tfinput = [tf.cast(inp, tf.float32) for inp in inputs]
                with tf.GradientTape() as tape:
                    tape.watch(tfinput)
                    pred = self.kerasmodel(tfinput)
                    if hasattr(outputs, '__len__') > 1:
                        total_output = K.sum([K.sum(o, axis=-1) for o in pred], axis=-1)
                    else:
                        total_output = K.sum(pred, axis=-1)

                grad = tape.gradient(total_output, tfinput)
                return [g.numpy() for g in grad]

            self._influence = _influence

        self.name = name

        self.outputdir = _get_output_root_directory()

        if not os.path.exists(self.outputdir):  # pragma: no cover
            # this is excluded from unit tests, because the testing
            # framework always provides a directory
            os.makedirs(self.outputdir)

        if not os.path.exists(os.path.join(self.outputdir, 'logs')):
            os.makedirs(os.path.join(self.outputdir, 'logs'))

        logfile = os.path.join(self.outputdir, 'logs', 'janggu.log')

        self.logger = logging.getLogger(self.name)

        logging.basicConfig(filename=logfile,
                            level=logging.DEBUG,
                            format='%(asctime)s:%(name)s:%(message)s',
                            datefmt='%m/%d/%Y %H:%M:%S')
        self.logger.info("Model Summary:")
        self.kerasmodel.summary(print_fn=self.logger.info)

    @classmethod
    def create_by_name(cls, name, custom_objects=None):
        """Creates a Janggu object by name.

        This option is used to load a pre-trained model.

        Parameters
        ----------
        name : str
            Name of the model.
        custom_objects : dict or None
            This allows loading of custom layers using load_model.
            All janggu specific layers are automatically included as custom_objects.
            Default: None

        Examples
        --------

        .. code-block:: python

          in_ = Input(shape=(10,), name='ip')
          layer = Dense(3)(in_)
          output = Dense(1, activation='sigmoid', name='out')(layer)

          # Instantiate a model.
          model = Janggu(inputs=in_, outputs=output, name='test_model')

          # saves the model to <janggu_results>/models
          model.save()

          # remove the original model
          del model

          # reload the model
          model = Janggu.create_by_name('test_model')
        """

        path = cls._storage_path(name, _get_output_root_directory())

        custom_objects = {} if not custom_objects else custom_objects

        custom_objects.update(JANGGU_LAYERS)

        model = load_model(path, custom_objects=custom_objects)
        return cls(model.inputs, model.outputs, name)

    @property
    def name(self):
        """Model name"""
        return self._name

    @name.setter
    def name(self, name):
        if not isinstance(name, str):
            raise Exception("Name must be a string.")
        self._name = name

    def save(self, filename=None, overwrite=True, show_shapes=True):
        """Saves the model.

        Parameters
        ----------
        filename : str
            Filename of the stored model. Default: None.
        overwrite: bool
            Overwrite a stored model. Default: True.
        """
        if not filename:  # pragma: no cover
            filename = self._storage_path(self.name, self.outputdir)

        plotname = os.path.splitext(filename)[0] + '.png'
        try:
            plot_model(self.kerasmodel, to_file=plotname, show_shapes=show_shapes)
        except Exception:  # pragma: no cover pylint: disable=broad-except
            # if graphviz is not installed on the system.
            self.logger.exception('plot_model failed, continue nevertheless: ')

        self.logger.info("Save model %s", filename)
        self.kerasmodel.save(filename, overwrite)

    def summary(self):
        """Prints the model definition."""
        self.kerasmodel.summary()

    @classmethod
    def create(cls, template, modelparams=None, inputs=None,
               outputs=None, name=None):
        """Janggu constructor method.

        This method instantiates a Janggu model with
        model template and parameters. It also allows to
        automatically infer and extend the correct
        input and output layers for the network.

        Parameters
        -----------
        template : function
            Python function that defines a model template of a neural network.
            The function signature must adhere to the signature
            :code:`template(inputs, inputs, outputs, modelparams)`
            and is expected to return
            :code:`(input_tensor, output_tensor)` of the neural network.
        modelparams : list or tuple or None
            Additional model parameters that are passed along to template
            upon creation of the neural network. For instance,
            this could contain number of neurons at each layer.
            Default: None.
        inputs : Dataset, list(Dataset) or None
            Input datasets from which the input layer shapes should be derived.
            Use this option together with the :code:`inputlayer` decorator (see Example below).
        outputs : Dataset, list(Dataset) or None
            Output datasets from which the output layer shapes should be derived.
            Use this option toghether with :code:`outputdense` or :code:`outputconv`
            decorators (see Example below).
        name : str or None
            Model name. If None, a unique model name is generated
            based on the model configuration and network architecture.

        Examples
        --------

        Specify a model using a model template and parameters:

        .. code-block:: python

          def test_manual_model(inputs, inp, oup, params):
              in_ = Input(shape=(10,), name='ip')
              layer = Dense(params)(in_)
              output = Dense(1, activation='sigmoid', name='out')(in_)
              return in_, output

          # Defines the same model by invoking the definition function
          # and the create constructor.
          model = Janggu.create(template=test_manual_model, modelparams=3)
          model.summary()

        Specify a model using automatic input and output layer determination.
        That is, only the model body needs to be specified:

        .. code-block:: python

          import numpy as np
          from janggu import Janggu
          from janggu import inputlayer, outputdense
          from janggu.data import Array

          # Some random data which you would like to use as input for the
          # model.
          DATA = Array('ip', np.random.random((1000, 10)))
          LABELS = Array('out', np.random.randint(2, size=(1000, 1)))

          # The decorators inputlayer and outputdense
          # extract the layer shapes and append the respective layers
          # to the network
          # so that only the model body remains to be specified.
          # Note that the the decorator order matters.
          # inputlayer must be specified before outputdense.
          @inputlayer
          @outputdense(activation='sigmoid')
          def test_inferred_model(inputs, inp, oup, params):
              with inputs.use('ip') as in_:
                  # the with block allows for easy
                  # access of a specific named input.
                  output = Dense(params)(in_)
              return in_, output

          # create a model.
          model = Janggu.create(template=test_inferred_model, modelparams=3,
                                name='test_model',
                                inputs=DATA,
                                outputs=LABELS)

        """

        model = create_model(template, modelparams, inputs, outputs)
        return cls(model.inputs, model.outputs, name)

    def compile(self, *args, **kargs):
                #loss_weights=None, sample_weight_mode=None,
                #weighted_metrics=None, target_tensors=None):
        """Model compilation.

        This method delegates the compilation to
        keras.models.Model.compile. See also https://keras.io/models/model/


        Examples
        --------

        .. code:: python

           model.compile(optimizer='adadelta', loss='binary_crossentropy')

        """

        self.kerasmodel.compile(*args, **kargs)

    def fit(self,  # pylint: disable=too-many-locals
            inputs=None,
            outputs=None,
            batch_size=None,
            epochs=1,
            verbose=1,
            callbacks=None,
            validation_data=None,
            shuffle=True,
            class_weight=None,
            sample_weight=None,
            initial_epoch=0,
            steps_per_epoch=None,
            use_multiprocessing=False,
            workers=1):
        """Model fitting.

        This method is used to fit a given model.
        Most of parameters are directly delegated the
        fit_generator of the keras model.

        Parameters
        ----------
        inputs : :code:`Dataset`, list(Dataset) or Sequence (keras.utils.Sequence)
            Input Dataset or Sequence to use for fitting the model.
        outputs : :code:`Dataset`, list(Dataset) or None
            Output Dataset containing the training targets. If a Sequence
            is used for inputs, outputs will have no effect.
        batch_size : int or None
            Batch size. If set to None a batch size of 32 is used.
        epochs : int
            Number of epochs. Default: 1.
        verbose : int
            Verbosity level. See https://keras.io.
        callbacks : List(keras.callbacks.Callback)
            Callbacks to be applied during training. See https://keras.io/callbacks
        validation_data : tuple, Sequence or None
            Validation data can be a tuple (input_dataset, output_dataset),
            or (input_dataset, output_dataset, sample_weights) or
            a keras.utils.Sequence instance or a list of validation chromsomoes.
            The latter choice only works with when using Cover and Bioseq dataset.
            This allows you to train on a dedicated set of chromosomes
            and to validate the performance on respective heldout chromosomes.
            If None, validation is not applied.
        shuffle : boolean
            shuffle batches. Default: True.
        class_weight : dict
            Class weights. See https://keras.io.
        sample_weight : np.array or None
            Sample weights. See https://keras.io.
        initial_epoch : int
            Initial epoch at which to start training.
        steps_per_epoch : int, None.
            Number of steps per epoch. If None, this value is determined from
            the dataset size and the batch_size.
        use_multiprocessing : boolean
            Whether to use multiprocessing. See https://keras.io. Default: False.
        workers : int
            Number of workers to use in multiprocessing mode. Default: 1.



        Examples
        --------

        .. code-block:: python

          model.fit(DATA, LABELS)

        """

        if not isinstance(inputs, Sequence):
            inputs = _convert_data(self.kerasmodel, inputs, 'input_layers')
            outputs = _convert_data(self.kerasmodel, outputs, 'output_layers')

        hyper_params = {
            'epochs': epochs,
            'batch_size': batch_size,
            'shuffle': shuffle,
            'class_weight': class_weight,
            'initial_epoch': initial_epoch,
            'steps_per_epoch': steps_per_epoch,
            'use_multiprocessing': use_multiprocessing,
            'workers': workers
        }

        self.logger.info('Fit: %s', self.name)
        if isinstance(inputs, Sequence):
            self.logger.info('using custom Sequence')
        else:
            self.logger.info("Input:")
            self.__dim_logging(inputs)
            self.logger.info("Output:")
            self.__dim_logging(outputs)
        self.timer = time.time()
        history = None
        self.logger.info("Hyper-parameters:")
        for par_ in hyper_params:
            self.logger.info('%s: %s', par_, str(hyper_params[par_]))

        callbacks = [] if callbacks is None else callbacks

        callbacks.append(LambdaCallback(on_epoch_end=lambda epoch, logs: self.logger.info(
            "epoch %s: %s",
            epoch + 1,
            ' '.join(["{}=".format(k) +
                      ('{:.4f}' if
                       abs(logs[k]) > 1e-3
                       else '{:.4e}').format(logs[k]) for k in logs]))))

        if not os.path.exists(os.path.join(self.outputdir, 'evaluation')):
            os.mkdir(os.path.join(self.outputdir, 'evaluation'))
        if not os.path.exists(os.path.join(self.outputdir, 'evaluation', self.name)):
            os.mkdir(os.path.join(self.outputdir, 'evaluation', self.name))

        callbacks.append(CSVLogger(os.path.join(self.outputdir,
                                                'evaluation',
                                                self.name,
                                                'training.log')))

        if not batch_size:
            batch_size = 32

        if isinstance(inputs, Sequence):
            # input could be a sequence
            jseq = inputs
        else:
            jseq = JangguSequence(inputs, outputs, sample_weight, batch_size, shuffle=shuffle)

        if isinstance(validation_data, tuple):
            valinputs = _convert_data(self.kerasmodel, validation_data[0],
                                      'input_layers')
            valoutputs = _convert_data(self.kerasmodel, validation_data[1],
                                       'output_layers')
            sweights = validation_data[2] if len(validation_data) == 3 else None
            valjseq = JangguSequence(valinputs, valoutputs, sweights, batch_size, shuffle=False)
        elif isinstance(validation_data, Sequence):
            valjseq = validation_data
        elif isinstance(validation_data, list) and isinstance(validation_data[0], str):
            # if the validation data is a list of chromosomes that should
            # be used as validation dataset we end up here.

            # This is only possible, however, if all input and output datasets
            # are Cover or Bioseq dataset.
            if not all(hasattr(datum, 'gindexer') \
                for datum in [jseq.inputs[k] for k in jseq.inputs] +
                       [jseq.outputs[k] for k in jseq.outputs]):
                raise ValueError("Not all dataset are Cover or Bioseq dataset"
                                 " which is required for this options.")

            # then split the original dataset into training and validation set.
            train, val = split_train_test((jseq.inputs, jseq.outputs), validation_data)

            traininp, trainoup = train
            valinp, valoup = val

            self.logger.info("Split in training and validation set:")
            self.logger.info("Training-Input:")
            self.__dim_logging(traininp)
            self.logger.info("Training-Output:")
            self.__dim_logging(trainoup)
            self.logger.info("Validation-Input:")
            self.__dim_logging(valinp)
            self.logger.info("Validation-Output:")
            self.__dim_logging(valoup)
            jseq = JangguSequence(_convert_data(self.kerasmodel, traininp,
                                                'input_layers'),
                                  _convert_data(self.kerasmodel, trainoup,
                                                'output_layers'),
                                  sample_weights=None, batch_size=jseq.batch_size, shuffle=jseq.shuffle)
            valjseq = JangguSequence(_convert_data(self.kerasmodel, valinp,
                                                   'input_layers'),
                                     _convert_data(self.kerasmodel, valoup,
                                                   'output_layers'),
                                     sample_weights=None, batch_size=jseq.batch_size, shuffle=False)

        else:
            valjseq = None


        try:
            history = self.kerasmodel.fit_generator(
                jseq,
                epochs=epochs,
                validation_data=valjseq,
                class_weight=class_weight,
                initial_epoch=initial_epoch,
                shuffle=shuffle,
                use_multiprocessing=use_multiprocessing,
                max_queue_size=50,
                workers=workers,
                verbose=verbose,
                callbacks=callbacks)
        except Exception:  # pragma: no cover
            # ignore the linter warning, the exception
            # is reraised anyways.
            self.logger.exception('fit_generator failed:')
            raise

        self.logger.info('#' * 40)
        for k in history.history:
            self.logger.info('%s: %f', k, history.history[k][-1])
        self.logger.info('#' * 40)

        self.save()
        self._save_hyper(hyper_params)

        self.logger.info("Training finished after %1.3f s",
                         time.time() - self.timer)
        return history

    def predict(self, inputs,  # pylint: disable=too-many-locals
                batch_size=None,
                verbose=0,
                steps=None,
                layername=None,
                datatags=None,
                callbacks=None,
                use_multiprocessing=False,
                workers=1):

        """Performs a prediction.

        This method predicts the targets.
        All of the parameters are directly delegated the
        predict_generator of the keras model.
        See https://keras.io/models/model/#methods.

        Parameters
        ----------
        inputs : :code:`Dataset`, list(Dataset) or Sequence (keras.utils.Sequence)
            Input Dataset or Sequence to use for fitting the model.
        batch_size : int or None
            Batch size. If set to None a batch size of 32 is used.
        verbose : int
            Verbosity level. See https://keras.io.
        steps : int, None.
            Number of predict steps. If None, this value is determined from
            the dataset size and the batch_size.
        layername : str or None
            Layername for which the prediction should be performed. If None,
            the output layer will be used automatically.
        datatags : list(str) or None
            Tags to annotate the evaluation results. Default: None.
        callbacks : List(:code:`Scorer`)
            Scorer instances to be applied on the predictions.
        use_multiprocessing : boolean
            Whether to use multiprocessing for the prediction. Default: False.
        workers : int
            Number of workers to use. Default: 1.


        Examples
        --------

        .. code-block:: python

          model.predict(DATA)

        """

        if not isinstance(inputs, Sequence):
            inputs = _convert_data(self.kerasmodel, inputs, 'input_layers')

        self.logger.info('Predict: %s', self.name)
        if isinstance(inputs, Sequence):
            self.logger.info('using custom Sequence')
        else:
            self.logger.info("Input:")
            self.__dim_logging(inputs)
        self.timer = time.time()

        # if a desired layername is specified, the features
        # will be predicted.
        if layername:
            model = Janggu(self.kerasmodel.input,
                           self.kerasmodel.get_layer(layername).output,
                           name=self.name)
        else:
            model = self

        if not batch_size:
            batch_size = 32

        if isinstance(inputs, Sequence):
            jseq = inputs
        else:
            jseq = JangguSequence(inputs, None, None, batch_size=batch_size)

        try:
            preds = model.kerasmodel.predict_generator(
                jseq,
                steps=steps,
                use_multiprocessing=use_multiprocessing,
                workers=workers,
                verbose=verbose)
        except Exception:  # pragma: no cover
            self.logger.exception('predict_generator failed:')
            raise

        prd = _convert_data(model.kerasmodel, preds, 'output_layers')
        if layername is not None:
            # no need to set an extra datatag.
            # if layername is present, it will be added to the tags
            if datatags is None:
                datatags = [layername]
            else:
                datatags.append(layername)
        for callback in callbacks or []:
            callback.score(model, prd, datatags=datatags)
        return preds

    def evaluate(self, inputs=None, outputs=None,  # pylint: disable=too-many-locals
                 batch_size=None,
                 sample_weight=None,
                 steps=None,
                 datatags=None,
                 callbacks=None,
                 use_multiprocessing=False,
                 workers=1):
        """Evaluates the performance.

        This method is used to evaluate a given model.
        All of the parameters are directly delegated the
        evalute_generator of the keras model.
        See https://keras.io/models/model/#methods.


        Parameters
        ----------
        inputs : :code:`Dataset`, list(Dataset) or Sequence (keras.utils.Sequence)
            Input Dataset or Sequence to use for evaluating the model.
        outputs :  :code:`Dataset`, list(Dataset) or None
            Output Dataset containing the training targets. If a Sequence
            is used for inputs, outputs will have no effect.
        batch_size : int or None
            Batch size. If set to None a batch size of 32 is used.
        sample_weight : np.array or None
            Sample weights. See https://keras.io.
        steps : int, None.
            Number of predict steps. If None, this value is determined from
            the dataset size and the batch_size.
        datatags : list(str) or None
            Tags to annotate the evaluation results. Default: None.
        callbacks : List(:code:`Scorer` or str)
            Scorer instances to be applied on the predictions. Furthermore,
            commonly used scoring metrics can be added by name, including
            'roc', 'auroc', 'prc', 'auprc' for evaluating binary classification
            applications and 'cor' (for Pearson's correlation), 'mae', 'mse'
            and 'var_explained' for regression applications.
        use_multiprocessing : boolean
            Whether to use multiprocessing for the prediction. Default: False.
        workers : int
            Number of workers to use. Default: 1.


        Examples
        --------

        .. code-block:: python

          model.evaluate(DATA, LABELS)

          # binary classification evaluation with callbacks
          model.evaluate(DATA, LABELS, callcacks=['auprc', 'auroc'])

        """

        self.logger.info('Evaluate: %s', self.name)
        if isinstance(inputs, Sequence):
            inputs_ = _convert_data(self.kerasmodel, inputs.inputs, 'input_layers')
            outputs_ = _convert_data(self.kerasmodel, inputs.outputs, 'output_layers')
            self.logger.info('Using custom Sequence.')
            self.logger.info("Input:")
            self.__dim_logging(inputs_)
            self.logger.info("Output:")
            self.__dim_logging(outputs_)
        else:
            inputs_ = _convert_data(self.kerasmodel, inputs, 'input_layers')
            outputs_ = _convert_data(self.kerasmodel, outputs, 'output_layers')
            self.logger.info("Input:")
            self.__dim_logging(inputs_)
            self.logger.info("Output:")
            self.__dim_logging(outputs_)
        self.timer = time.time()

        if not batch_size:
            batch_size = 32

        if isinstance(inputs, Sequence):
            jseq = inputs
        else:
            jseq = JangguSequence(inputs_, outputs_, sample_weight, batch_size=batch_size)

        try:
            values = self.kerasmodel.evaluate_generator(
                jseq,
                steps=steps,
                use_multiprocessing=use_multiprocessing,
                workers=workers)
        except Exception:  # pragma: no cover
            self.logger.exception('evaluate_generator failed:')
            raise

        self.logger.info('#' * 40)
        values = _to_list(values)

        for i, value in enumerate(values):
            self.logger.info('%s: %f', self.kerasmodel.metrics_names[i], value)
        self.logger.info('#' * 40)

        self.logger.info("Evaluation finished in %1.3f s",
                         time.time() - self.timer)

        preds = self.kerasmodel.predict_generator(jseq, steps=steps,
                                                  use_multiprocessing=use_multiprocessing,
                                                  workers=workers)
        preds = _convert_data(self.kerasmodel, preds, 'output_layers')

        for callback in callbacks or []:
            callback = get_scorer(callback)
            callback.score(self, preds, outputs=outputs_, datatags=datatags)
        return values

    def predict_variant_effect(self,  # pylint: disable=too-many-locals
                               bioseq,
                               variants,
                               conditions,
                               output_folder,
                               condition_filter=None,
                               batch_size=None,
                               annotation=None,
                               ignore_reference_match=False,
                               order=1):
        """Evaluates the performance.

        see help(predict_variant_effect) for a description of
        the parameters

        Examples
        --------

        .. code-block:: python

          # Evaluate all variants and all conditions (outputs)
          model.predict_variant_effect(DATA, VARIANTS, CONDITIONS,
                                       'vcfoutput')

          # Evaluate all variants and a subset of conditions (Ctcf output labels)
          model.predict_variant_effect(DATA, LABELS, CONDITIONS,
                                       'vcfoutput_subset',
                                       contition_filter='Cfcf')

        """
        return predict_variant_effect(self.kerasmodel,
                                      bioseq,
                                      variants,
                                      conditions,
                                      output_folder,
                                      condition_filter=condition_filter,
                                      batch_size=batch_size,
                                      annotation=annotation,
                                      ignore_reference_match=ignore_reference_match)

    def __dim_logging(self, data):
        if isinstance(data, dict):
            for key in data:
                self.logger.info("\t%s: %s", key, data[key].shape)

        if hasattr(data, "shape"):
            data = [data]

        if isinstance(data, list):
            for datum in data:
                self.logger.info("\t%s", datum.shape)

    def get_config(self):
        """Get model config."""
        return self.kerasmodel.get_config()

    @staticmethod
    def _storage_path(name, outputdir):
        """Returns the path to the model storage file."""
        if not os.path.exists(os.path.join(outputdir, "models")):
            os.mkdir(os.path.join(outputdir, "models"))
        filename = os.path.join(outputdir, 'models', '{}.h5'.format(name))
        return filename

    def _save_hyper(self, hyper_params, filename=None):
        """This method attaches the hyper parameters that were used
        to train the model to the hdf5 file.

        This method is supposed to be used after model training
        from within the fit method.
        It attaches the hyper parameter, e.g. epochs, batch_size, etc.
        to the hdf5 file that contains the model weights.
        The hyper parameters are added as attributes to the
        group 'model_weights'.

        Parameters
        ----------
        hyper_parameters : dict
            Dictionary that contains the hyper parameters.
        filename : str
            Filename of the hdf5 file. This file must already exist.
            So save has to be called before.
        """
        if not filename:  # pragma: no cover
            filename = self._storage_path(self.name, self.outputdir)

        content = h5py.File(filename, 'r+')
        weights = content['model_weights']
        for key in hyper_params:
            if hyper_params[key]:
                weights.attrs[key] = str(hyper_params[key])
        content.close()


def model_from_json(json_string, custom_objects=None):
    """Wrapping keras.models.model_from_json"""
    custom_objects = {} if custom_objects is None else custom_objects

    custom_objects.update(JANGGU_LAYERS)

    return keras.models.model_from_json(json_string, custom_objects)


def model_from_yaml(yaml_string, custom_objects=None):
    """Wrapping keras.models.model_from_yaml"""
    custom_objects = {} if custom_objects is None else custom_objects

    custom_objects.update(JANGGU_LAYERS)

    return keras.models.model_from_yaml(yaml_string, custom_objects)


def create_model(template, modelparams=None, inputs=None,
                 outputs=None, name=None):
    """Constructor method.

    This method instantiates a keras model with
    model template and parameters. It also allows to
    automatically infer and extend the correct
    input and output layers for the network.

    Parameters
    -----------
    template : function
        Python function that defines a model template of a neural network.
        The function signature must adhere to the signature
        :code:`template(inputs, inputp, outputp, modelparams)`
        and is expected to return
        :code:`(input_tensor, output_tensor)` of the neural network.
    modelparams : list or tuple or None
        Additional model parameters that are passed along to template
        upon creation of the neural network. For instance,
        this could contain number of neurons at each layer.
        Default: None.
    inputs : Dataset, list(Dataset) or None
        Input datasets from which the input layer shapes should be derived.
        Use this option together with the :code:`inputlayer` decorator (see Example below).
    outputs : Dataset, list(Dataset) or None
        Output datasets from which the output layer shapes should be derived.
        Use this option toghether with :code:`outputdense` or :code:`outputconv`
        decorators (see Example below).
    name : str or None
        Model name. If None, a unique model name is generated
        based on the model configuration and network architecture.

    Examples
    --------

    Specify a model using a model template and parameters:

    .. code-block:: python

      def test_manual_model(inputs, inp, oup, params):
          in_ = Input(shape=(10,), name='ip')
          layer = Dense(params)(in_)
          output = Dense(1, activation='sigmoid', name='out')(in_)
          return in_, output

      # Defines the same model by invoking the definition function
      # and the create constructor.
      model = create_model(template=test_manual_model, modelparams=3)
      model.summary()

    Specify a model using automatic input and output layer determination.
    That is, only the model body needs to be specified:

    .. code-block:: python

      import numpy as np
      from janggu import inputlayer, outputdense
      from janggu.data import Array

      # Some random data which you would like to use as input for the
      # model.
      DATA = Array('ip', np.random.random((1000, 10)))
      LABELS = Array('out', np.random.randint(2, size=(1000, 1)))

      # The decorators inputlayer and outputdense
      # extract the layer shapes and append the respective layers
      # to the network
      # so that only the model body remains to be specified.
      # Note that the the decorator order matters.
      # inputlayer must be specified before outputdense.
      @inputlayer
      @outputdense(activation='sigmoid')
      def test_inferred_model(inputs, inp, oup, params):
          with inputs.use('ip') as in_:
              # the with block allows for easy
              # access of a specific named input.
              output = Dense(params)(in_)
          return in_, output

      # create a model.
      model = create_model(template=test_inferred_model, modelparams=3,
                            name='test_model',
                            inputp=DATA,
                            outputp=LABELS)

    """

    inputs_ = _data_props(inputs) if inputs else None
    outputs_ = _data_props(outputs) if outputs else None

    input_tensors, output_tensors = template(None, inputs_,
                                             outputs_, modelparams)

    model = Model(inputs=input_tensors, outputs=output_tensors, name=name)

    return model


def input_attribution(model, inputs,  # pylint: disable=too-many-locals
                      chrom=None, start=None, end=None,
                      idx=None):

    """Evaluates the integrated gradients method on the input coverage tracks.

    This allows to attribute feature importance values to the prediction scores.
    Integrated gradients have been introduced in
    Sundararajan, Taly and Yan, Axiomatic Attribution for Deep Networks.
    PMLR 70, 2017.

    The method can either be called, by specifying the region of interest directly
    by setting chrom, start and end. Alternatively, it is possible to specify the
    region index. For example, the n^th region of the dataset.

    Parameters
    ----------
    model : Janggu
        Janggu model wrapper
    inputs : :code:`Dataset`, list(Dataset)
        Input Dataset.
    chrom : str or None
        Chromosome name.
    start : int or None
        Region start.
    end : int or None
        Region end.
    idx : int or None
        The index of the i^th sequence to use for the attribution.
        If idx is set chrom, start and end are ignored.

    Examples
    --------

    .. code-block:: python

      # Suppose DATA is a Bioseq or Cover object
      # To query the input feature importance of a specific genomic region
      # use
      input_attribution(model, DATA, chrom='chr1', start=start, end=end)

      # To query the input feature importance using a specific index
      input_attribution(model, DATA, idx=0)
    """

    inputs = _to_list(inputs)
    if idx is not None:
        iv = inputs[0].gindexer[idx]
        chrom, start, end = iv.chrom, iv.start, iv.end

    output_chrom, output_start, output_end = chrom, start, end


    # store original gindexer
    gindexers_save = [ip.gindexer for ip in inputs]

    # create new indexers ranging only over the selected region
    # if chrom, start, end was supplied retrieve the respective indices
    index_list = [gi.idx_by_region(include=output_chrom,
                                   start=output_start,
                                   end=output_end) for gi in gindexers_save]

    # first construct the union of indices
    index_set = set()
    for idx_list_el in index_list:
        index_set = index_set | set(idx_list_el)

    # only keep the indices that remain in the across all inputs
    # indices that are only present in some of the inputs are discarded.
    for idx_list_el in index_list:
        index_set = index_set & set(idx_list_el)

    idxs = list(index_set)
    idxs.sort()

    subgindexers = [copy.copy(gi) for gi in gindexers_save]
    for subgi in subgindexers:
        subgi.chrs = [subgi.chrs[i] for i in idxs]
        subgi.starts = [subgi.starts[i] for i in idxs]
        subgi.ends = [subgi.ends[i] for i in idxs]
        subgi.strand = [subgi.strand[i] for i in idxs]

    # assign it to the input datasets temporarily
    for inp, _ in enumerate(inputs):
        inputs[inp].gindexer = subgindexers[inp]

    try:
        #allocate arrays
        output = [np.zeros((1, output_end-output_start,
                            inp.shape[-2], inp.shape[-1])) for inp in inputs]
        resols = [inp.garray.resolution for inp in inputs]

        for igi in range(len(inputs[0])):

            # current influence
            influence = [np.zeros((1,) + inp.shape[1:]) for inp in inputs]

            # get influence for current window with integrated gradient
            x_in = [inp[igi] for inp in inputs]
            for step in range(1, 51):
                grad = model._influence([x*step/50 for x in x_in])
                for iinp, inp in enumerate(x_in):
                    for idim, _ in np.ndenumerate(inp):
                        influence[iinp][idim] += (x_in[iinp][idim]/50)*grad[iinp][idim]

            # scale length to nucleotide resolution
            influence = [np.repeat(influence[i], resols[i],
                                   axis=1) for i, _ in enumerate(inputs)]

            for iout in range(len(output)):
                if influence[iout].shape[1] < inputs[iout].gindexer[igi].length:
                    order = inputs[iout].gindexer[igi].length - influence[iout].shape[1]
                else:
                    order = 0
                # incremetally add the influence results into the output
                # array for all subwindows in the genomic indexer

                if output_start < inputs[iout].gindexer[igi].start:
                    ostart = inputs[iout].gindexer[igi].start - output_start
                    lstart = 0
                else:
                    ostart = 0
                    lstart = output_start - inputs[iout].gindexer[igi].start

                if output_end > inputs[iout].gindexer[igi].end:
                    oend = inputs[iout].gindexer[igi].end - output_start
                    lend = inputs[iout].gindexer[igi].end - inputs[iout].gindexer[igi].start
                else:
                    oend = output_end - output_start
                    lend = output_end - inputs[iout].gindexer[igi].start

                # for mutually overlapping positions, we employ a heuristic
                # that keeps the maximum influence over the overlapping intervals
                # spanning the position
                m = np.zeros((2,) + (1, inputs[iout].gindexer[igi].length, ) \
                             + influence[iout].shape[2:], dtype=influence[iout].dtype)

                m[0][:, lstart:lend, :, :] = output[iout][:, (ostart):(oend), :, :]
                m[1][:, lstart:(lend - order), :, :] = \
                    influence[iout][:, lstart:(lend - order), :, :]
                m = np.abs(m).max(axis=0)
                m = m[:, lstart:lend, :, :]
                output[iout][:, ostart:oend, :, :] = m

        for iout in range(len(output)):
            # finally wrap the output up as coverage track
            output[iout] = Cover.create_from_array('attr_'+inputs[iout].name,
                                                   output[iout],
                                                   GenomicIndexer.create_from_region(
                                                       chrom, start, end, '.',
                                                       binsize=end-start,
                                                       stepsize=1, flank=0),
                                                   conditions=inputs[iout].conditions)

        for inp, _ in enumerate(inputs):
            # restore the initial genomic indexers
            inputs[inp].gindexer = gindexers_save[inp]

    except Exception:  # pragma: no cover
        model.logger.exception('_influence failed:')
        raise

    return output


def predict_variant_effect(model,  # pylint: disable=too-many-locals
                           bioseq,
                           variants,
                           conditions,
                           output_folder,
                           condition_filter=None,
                           batch_size=None,
                           annotation=None,
                           ignore_reference_match=False,
                           order=1):
    """Evaluates the performance.

    Parameters
    ----------
    model : :code:`keras.Model`
        A keras model
    bioseq : :code:`Bioseq` or str
        Bioseq container containing a reference genome or the reference genome file
        in fasta format.
        If a Bioseq object is used, it has to be loaded with store_whole_genome=True.
        This may be faster for large amounts of variants.
        Consider using cache=True as well, such that the genome only needs to be loaded
        once, and reloaded from the cache if needed.
        Alternatively, a string pointing to the reference genome sequence in FASTA
        format can be supplied from which the sequence context is extracted.
    variants :  str
        File name of a VCF file containg the variants under study.
    conditions : list(str)
        Condition labels for each output prediction.
    output_folder : str
        The method produces an hdf5 and a bed file as output.
        The bed-file contains the variant positions while the
        hdf5 file contains the reference and alternative variant scores
        for each output feature.
    condition_filter : str or None
        Regular expression filter on which conditions should be evaluated.
        If None, all output conditions will be returned.
    batch_size : int, None.
        Batch size. If None, a batch_size of 128 is used.
    annotation : BedTool object or None
        BedTool holding feature annotation e.g. gene annotation.
        The annotation may be used to perform strand-specific variant effect
        predictions. Each variant is intersected with the annotation
        in order to derive the correct strandedness. If variants
        do not overlap with an annotation features or for missing annotation,
        the forward strand is used.
    ignore_reference_match : boolean
        Whether to ignore mismatches between the reference sequence and
        the reference base in the VCF file. If False, the variant will
        be skipped over and only matching positions are processed.
        Otherwise all variants will be processed. Default: False.
    order : int
        Order of the DNA sequence representation. Only relevant if the option
        bioseq is pointing to a reference genome file. If bioseq is already a
        Bioseq object, this options will be ignored. Default: 1.

    Returns
    -------
    tuple:
        Tuple containing the output filenames: an hdf5 and a bed file.

    Examples
    --------

    .. code-block:: python

      # Evaluate all variants and all conditions (outputs)
      predict_variant_effect(kerasmodel, DATA, VARIANTS, CONDITIONS,
                                   'vcfoutput')

      # Evaluate all variants and a subset of conditions (Ctcf output labels)
      predict_variant_effect(kerasmodel, DATA, LABELS, CONDITIONS,
                                   'vcfoutput_subset',
                                   contition_filter='Cfcf')

    """
    if batch_size is None:
        batch_size = 128

    def _get_input_layer_shape(layer):
        if isinstance(layer.input_shape, list):
            return layer.input_shape[0]
        return layer.input_shape

    if isinstance(bioseq, Bioseq):
        order = bioseq.garray.order

    if len(model.inputs) > 1:
        raise ValueError('Only one input layer supported for predict_variant_effect.')
    binsize = _get_input_layer_shape(model.layers[0])[1] + order - 1

    # the network might output arbitrarily many
    # output.
    # With the filter option it is possible to
    # restrict the analysis to certain features.
    if condition_filter is None:
        conditions = [(idx, cond) for idx, cond in enumerate(conditions)]
    else:
        conditions = [(idx, cond) for idx, cond in enumerate(conditions) \
                      if hasattr(re.search(condition_filter, cond), 'start')]


    icond = [el[0] for el in conditions]

    local_model = model

    if len(conditions) != model.output_shape[-1]:
        raise ValueError("The number of conditions does not match with the "
                         "number of network output units. [{}!={}]".format(
                             len(conditions),
                             model.output_shape[-1]))

    # get number of variants
    variantsstream = VariantStreamer(bioseq, variants, binsize, batch_size,
                                     annotation=annotation,
                                     ignore_reference_match=ignore_reference_match,
                                     order=order)

    nvariants = variantsstream.get_variant_count()

    h5file = h5py.File(os.path.join(output_folder, 'scores.hdf5'), 'w')

    h5file.create_dataset('labels', (len(conditions),),
                          dtype=h5py.special_dtype(vlen=str),
                          data=np.array([c[-1] for c in conditions],
                                        dtype=h5py.special_dtype(vlen=str)))

    refscore = h5file.create_dataset('refscore', (nvariants, len(conditions)),
                                     dtype='float16')
    altscore = h5file.create_dataset('altscore', (nvariants, len(conditions)),
                                     dtype='float16')
    diffscore = h5file.create_dataset('diffscore', (nvariants, len(conditions)),
                                      dtype='float16')
    logodds = h5file.create_dataset('logoddsscore', (nvariants, len(conditions)),
                                    dtype='float16')

    bar = Bar('Parsing {}: '.format(variants), max=int(np.ceil(nvariants/float(batch_size))))

    chromlist = []
    poslist = []
    vnamelist = []
    reflist = []
    altlist = []
    ibatch = 0

    # read variants file
    for names, chroms, poss, ra, aa, reference, alternative in variantsstream.flow():
        bar.next()

        if reference.shape[0] <= 0:
            # reached the end of the file
            break

        ref_score = local_model.predict_on_batch(reference)
        alt_score = local_model.predict_on_batch(alternative)

        chromlist += chroms
        poslist += poss
        vnamelist += names
        reflist += ra
        altlist += aa

        refscore[ibatch:(ibatch + ref_score.shape[0])] = ref_score[:, icond].astype('float16')
        altscore[ibatch:(ibatch + ref_score.shape[0])] = alt_score[:, icond].astype('float16')
        diffscore[ibatch:(ibatch + ref_score.shape[0])] = \
            alt_score[:, icond].astype('float16') - ref_score[:, icond].astype('float16')
        logodds[ibatch:(ibatch + ref_score.shape[0])] = \
            np.log(alt_score[:, icond].astype('float16')/
                   ref_score[:, icond].astype('float16') + 1e-7)
        ibatch += ref_score.shape[0]

    #form large string
    BedTool('\n'.join(['{} {} {} {}_{}>{}'.format(chrom, start, start+1, name, ref, alt)
                       for chrom, start, name, ref, alt in zip(chromlist,
                                                               poslist,
                                                               vnamelist, reflist, altlist)]),
            from_string=True).saveas(os.path.join(output_folder, 'snps.bed.gz'))

    bar.finish()
    h5file.close()

    return (os.path.join(output_folder, 'scores.hdf5'),
            os.path.join(output_folder, 'snps.bed.gz'))




import os

import matplotlib
import pkg_resources
import pytest

from janggu.data import Bioseq
from janggu.data import split_train_test
from janggu.data import subset
from janggu.data import view
from janggu.data.data import _data_props

matplotlib.use('AGG')


def test_dna_props_extraction(tmpdir):
    os.environ['JANGGU_OUTPUT'] = tmpdir.strpath
    data_path = pkg_resources.resource_filename('janggu', 'resources/')
    bed_file = os.path.join(data_path, 'sample.bed')

    refgenome = os.path.join(data_path, 'sample_genome.fa')

    dna = Bioseq.create_from_refgenome('dna', refgenome=refgenome,
                                    storage='ndarray',
                                    roi=bed_file,
                                    binsize=200, stepsize=200,
                                    order=1)

    props = _data_props(dna)
    assert 'dna' in props
    assert props['dna']['shape'] == (200, 1, 4)

    with pytest.raises(Exception):
        _data_props((0,))


def test_split_train_test():
    data_path = pkg_resources.resource_filename('janggu', 'resources/')
    bed_file = os.path.join(data_path, 'sample.bed')

    refgenome = os.path.join(data_path, 'sample_genome.fa')

    dna = Bioseq.create_from_refgenome('dna', refgenome=refgenome,
                                       storage='ndarray',
                                       roi=bed_file,
                                       binsize=200, stepsize=200,
                                       order=1, store_whole_genome=True)

    traindna, testdna = split_train_test(dna, holdout_chroms='chr2')

    assert len(traindna) == 50
    assert len(testdna) == 50
    assert len(dna) == len(traindna) + len(testdna)

    traindna, testdna = split_train_test([dna, dna], holdout_chroms='chr2')

    assert len(traindna[0]) == 50
    assert len(testdna[0]) == 50
    assert len(dna) == len(traindna[0]) + len(testdna[0])


def test_subset_include_chrname_test():
    data_path = pkg_resources.resource_filename('janggu', 'resources/')
    bed_file = os.path.join(data_path, 'sample.bed')

    refgenome = os.path.join(data_path, 'sample_genome.fa')

    dna = Bioseq.create_from_refgenome('dna', refgenome=refgenome,
                                       storage='ndarray',
                                       roi=bed_file,
                                       binsize=200, stepsize=200,
                                       order=1, store_whole_genome=True)

    subdna = subset(dna, include_regions='chr2')

    assert len(subdna) == 50


def test_subset_exclude_chrname_test():
    data_path = pkg_resources.resource_filename('janggu', 'resources/')
    bed_file = os.path.join(data_path, 'sample.bed')

    refgenome = os.path.join(data_path, 'sample_genome.fa')

    dna = Bioseq.create_from_refgenome('dna', refgenome=refgenome,
                                       storage='ndarray',
                                       roi=bed_file,
                                       binsize=200, stepsize=200,
                                       order=1, store_whole_genome=True)

    subdna = subset(dna, exclude_regions='chr2')

    assert len(subdna) == 50


def test_view_bed_test():
    data_path = pkg_resources.resource_filename('janggu', 'resources/')
    bed_file = os.path.join(data_path, 'sample.bed')
    bedsub_file = os.path.join(data_path, 'scored_sample.bed')

    refgenome = os.path.join(data_path, 'sample_genome.fa')

    dna = Bioseq.create_from_refgenome('dna', refgenome=refgenome,
                                       storage='ndarray',
                                       roi=bed_file,
                                       binsize=200, stepsize=200,
                                       order=1, store_whole_genome=True)

    subdna = view(dna, use_regions=bedsub_file)

    assert len(subdna) == 4



"""Model evaluation utilities.

This module contains classes and methods for simplifying
model evaluation.
"""

import datetime
import logging
import os

import numpy
from sklearn.metrics import average_precision_score
from sklearn.metrics import explained_variance_score
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import mean_squared_error
from sklearn.metrics import precision_recall_curve
from sklearn.metrics import roc_auc_score
from sklearn.metrics import roc_curve

from janggu.utils import ExportJson
from janggu.utils import ExportScorePlot
from janggu.utils import ExportTsv
from janggu.utils import _to_list


def _dimension_match(kerasmodel, data, layertype):
    """Check if layer dimensions match.
    The function checks whether the kerasmodel as compatible with
    the supplied inputs.

    Parameters
    ----------
    kerasmodel : :class:`keras.Model`
        Object of type keras.Model.
    data : Dataset or list(Dataset)
        Dataset to check compatiblity for.
    layertype : str
        layers is either 'input_layers' or 'output_layers'.

    Returns
    -------
    boolean :
        Returns True if the keras model is compatible with the data
        and otherwise False.
    """
    if data is None and layertype == 'output_layers':
        return True

    tmpdata = _to_list(data)

    if len(kerasmodel.get_config()[layertype]) != len(tmpdata):
        return False
    # Check if output dims match between model spec and data
    for datum in tmpdata:

        if datum.name not in [el[0] for el in
                              kerasmodel.get_config()[layertype]]:
            # If the layer name is not present we end up here
            return False
        layer = kerasmodel.get_layer(datum.name)
        oshape = layer.output_shape
        if isinstance(oshape, list):
            # this case is required for keras 2.4.3 and tf 2
            # which returns a list of tuples
            oshape = oshape[0]
        if not oshape[1:] == datum.shape[1:]:
            # if the layer name is present but the dimensions
            # are incorrect, we end up here.
            return False
    return True


def _reshape(data, percondition):
    """Reshape the dataset to make it compatible with the
    evaluation method.

    Parameters
    ----------
    data : dict(Dataset)
        A dictionary of datasets
    percondition : boolean
        Indicates whether to keep the condition (last) dimension or flatten
        over the condition.
    """

    if isinstance(data, dict):
        if percondition:
            # currently this only works for channel_last
            data = {k: data[k][:].reshape(
                (int(numpy.prod(data[k].shape[:-1])),
                 data[k].shape[-1])) for k in data}
        else:
            data = {k: data[k][:].reshape(
                (numpy.prod(data[k].shape[:]), 1)) for k in data}
    else:
        raise ValueError('Data must be a dict not {}'.format(type(data)))

    return data


class Scorer(object):
    """Scorer class.

    This class implements the callback interface that is used
    with :code:`Janggu.evaluate` and :code:`Janggu.predict`.
    The scorer maintains a scoring callable and an exporter callable
    which take care of determining the desired score and writing
    the result into a desired file, e.g. json, tsv or a figure, respectively.


    Parameters
    ----------
    name : str
        Name of the score to be performed.
    score_fct : None or callable
        Callable that is invoked for scoring.
        This callable must satisfy the signature
        :code:`fct(y_true, y_pred)` if used with
        :code:`Janggu.evaluate` and :code:`fct(y_pred)` if
        used with :code:`Janggu.predict`. The returned score should be
        compatible with the exporter.
    conditions : list(str) or None
        List of strings describing the conditions dimension of the dataset
        that is processed. If None, conditions are extracted from the
        y_true Dataset, if available. Otherwise, the conditions are integers
        ranging from zero to :code:`len(conditions) - 1`.
    exporter : callable
        Exporter function is used to export the scoring results
        in the desired manner,
        e.g. as json or tsv file. This function must satisfy the signature
        :code:`fct(output_path, filename_prefix, results)`.
    immediate_export : boolean
        If set to True, the exporter function will be invoked immediately
        after the evaluation of the dataset. If set to False, the results
        are maintained in memory which allows to export the results as a
        collection rather than individually.
    percondition : boolean
        Indicates whether the evaluation should be performed per condition
        or across all conditions. The former determines a score for each
        output condition, while the latter first flattens the array and then
        scores across conditions. Default: percondition=True.
    subdir : str
        Name of the subdir to store the output in. Default: None
        means the results are stored in the 'evaluation' subdir.
    """

    def __init__(self, name, score_fct=None,
                 conditions=None,
                 exporter=ExportJson(),
                 immediate_export=True,
                 percondition=True,
                 subdir=None):
        # append the path by a folder 'AUC'
        self.score_name = name
        self.score_fct = score_fct
        self.percondition = percondition
        self.logger = logging.getLogger('scorer')

        self.results = dict()
        self._exporter = exporter

        self.immediate_export = immediate_export
        self.conditions = conditions
        if subdir is None:
            subdir = 'evaluation'
        self.subdir = subdir

    def export(self, path, collection_name, datatags=None):
        """Exporting of the results.

        When calling export, the results which have been collected
        in self.results by using the score method are
        written to disk by invoking the supplied exporter function.

        Parameters
        ----------
        path : str
            Output directory.
        collection_name : str
            Subdirectory in which the results should be stored. E.g. Modelname.
        datatags : list(str) or None
            Optional tags describing the dataset. E.g. 'training_set'.
            Default: None
        """
        output_path = os.path.join(path, collection_name)
        if datatags is not None:
            output_path = os.path.join(output_path, '-'.join(datatags))
        if not os.path.exists(output_path):
            os.makedirs(output_path)

        if self.results:
            # if there are some results, export them
            self.logger.info(' '.join(('exporting', self.score_name, 'to', output_path)))
            self._exporter(output_path, self.score_name,
                           self.results)

    def score(self, model, predicted, outputs=None, datatags=None):
        """Scoring of the predictions relative to true outputs.

        When calling score, the provided
        score_fct is applied
        for each layer and condition separately.
        The result scores are maintained in a dict that uses
        :code:`(modelname, layername, conditionname)` as key
        and as values another dict of the form:
        :code:`{'date':<currenttime>, 'value': derived_score, 'tags':datatags}`.

        Parameters
        ----------
        model : :class:`Janggu`
            a Janggu object representing the current model.
        predicted: dict{name: np.array}
            Predicted outputs.
        outputs : dict{name: Dataset} or None
            True output labels. The Scorer is used with :code:`Janggu.evaluate`
            this argument will be present. With :code:`Janggu.evaluate` it is
            absent.
        datatags : list(str) or None
            Optional tags describing the dataset, e.g. 'test_set'.
        """

        if not datatags:
            datatags = []

        if outputs is not None:
            _out = _reshape(outputs, self.percondition)
        _pre = _reshape(predicted, self.percondition)
        self.logger.info(' '.join(('scoring:', self.score_name)))
        score_fct = self.score_fct
        if score_fct is None and outputs is not None:
            raise ValueError('Scorer: without outputs a score_fct must be supplied.')

        if score_fct is None:
            def _dummy(value):
                return value
            score_fct = _dummy

        for layername in model.get_config()['output_layers']:

            for idx in range(_pre[layername[0]].shape[-1]):

                if outputs is None:
                    score = score_fct(_pre[layername[0]][:, idx])
                else:
                    score = score_fct(_out[layername[0]][:, idx],
                                      _pre[layername[0]][:, idx])

                if not self.percondition:
                    condition = 'across'
                elif self.conditions is not None and \
                   len(self.conditions) == _pre[layername[0]].shape[-1]:
                    # conditions were supplied manually
                    condition = self.conditions[idx]
                elif outputs is not None and hasattr(outputs[layername[0]],
                                                     "conditions"):
                    # conditions are extracted from the outputs dataset
                    condition = outputs[layername[0]].conditions[idx]
                else:
                    # not conditions present, just number them.
                    condition = str(idx)

                try:
                    iter(score)
                except TypeError:
                    # if the score is a scalar value, we write it into
                    # the log file.
                    self.logger.info(' '.join((self.score_name,
                                               model.name,
                                               layername[0],
                                               condition,
                                               ":", str(score))))

                key = (condition,)
                if len(model.get_config()['output_layers']) > 1:
                    key = (layername[0],) + key
                if not self.immediate_export:
                    key = (model.name,) + key
                self.results[key] = \
                    {'date': str(datetime.datetime.utcnow()),
                     'value': score}

        if self.immediate_export:
            # export directly if required
            output_dir = os.path.join(model.outputdir, self.subdir)

            self.export(output_dir, model.name, datatags)

            # reset the results
            self.results = {}


# some standard evaluations are provided directly

# evaluation metrics from sklearn.metrics
def wrap_roc_(y_true, y_pred):
    """Helper function to determine the ROC.
    """
    fpr, tpr, _ = roc_curve(y_true, y_pred)
    aux = str('({:.2%})'.format(roc_auc_score(y_true, y_pred)))
    return fpr, tpr, aux


def wrap_prc_(y_true, y_pred):
    """Helper function to determine the PRC"""
    precision, recall, _ = precision_recall_curve(y_true, y_pred)
    aux = str('({:.2%})'.format(average_precision_score(y_true, y_pred)))
    return recall, precision, aux

def wrap_cor_(y_true, y_pred):
    """Helper function to determine the Pearson's correlation coeff."""
    return numpy.corrcoef(y_true, y_pred)[0, 1]


def get_scorer(scorer):
    """Function maps string names to the Scorer objects.

    This function takes a scorer by name or a Scorer object
    and returns an instantiation of a Scorer object.
    """
    if isinstance(scorer, Scorer):
        pass
    elif scorer in ['ROC', 'roc']:
        scorer = Scorer(scorer, wrap_roc_,
                        exporter=ExportScorePlot(xlabel='FPR', ylabel='TPR'))
    elif scorer in ['PRC', 'prc']:
        scorer = Scorer(scorer, wrap_prc_,
                        exporter=ExportScorePlot(xlabel='Recall',
                                                 ylabel='Precision'))
    elif scorer in ['auc', 'AUC', 'auROC', 'auroc']:
        scorer = Scorer(scorer, roc_auc_score, exporter=ExportTsv())
    elif scorer in ['auprc', 'auPRC', 'ap', 'AP']:
        scorer = Scorer(scorer, average_precision_score, exporter=ExportTsv())
    elif scorer in ['cor', 'pearson']:
        scorer = Scorer(scorer, wrap_cor_, exporter=ExportTsv())
    elif scorer in ['var_explained']:
        scorer = Scorer(scorer, explained_variance_score, exporter=ExportTsv())
    elif scorer in ['mse', 'MSE']:
        scorer = Scorer(scorer, mean_squared_error, exporter=ExportTsv())
    elif scorer in ['mae', 'MAE']:
        scorer = Scorer(scorer, mean_absolute_error, exporter=ExportTsv())
    else:
        raise ValueError("scoring callback {} unknown.".format(scorer))
    return scorer



"""Genomic arrays"""

import hashlib
import os
from collections import OrderedDict

import h5py
import numpy as np
from pybedtools import Interval
from scipy import sparse

from janggu.utils import _get_output_data_location
from janggu.utils import _iv_to_str
from janggu.utils import _str_to_iv


def _get_iv_length(length, resolution):
    """obtain the chromosome length for a given resolution."""
    if resolution is None:
        return 1
    return int(np.ceil(float(length)/resolution))


def get_collapser(method):
    """Get collapse method."""

    if method is None:
        return None
    elif  callable(method):
        return method
    elif method == 'mean':
        return lambda x: x.mean(axis=1)
    elif method == 'sum':
        return lambda x: x.sum(axis=1)
    elif method == 'max':
        return lambda x: x.max(axis=1)

    raise ValueError('Unknown method: {}'.format(method))


def create_sha256_cache(data, parameters):
    """Cache file determined from files and parameters."""

    sha256_hash = hashlib.sha256()

    # add file content to hash
    for datum in data or []:
        if isinstance(datum, str) and os.path.exists(datum):
            with open(datum, 'rb') as file_:
                for bblock in iter(lambda: file_.read(1024**2), b""):
                    sha256_hash.update(bblock)
        elif isinstance(datum, np.ndarray):
            sha256_hash.update(datum.tobytes())
        else:
            sha256_hash.update(str(datum).encode('utf-8'))

    # add parameter settings to hash
    sha256_hash.update(str(parameters).encode('utf-8'))

    return sha256_hash.hexdigest()


def _get_cachefile(cachestr, tags, fileending):
    """ Determine cache file location """
    filename = None
    if cachestr is not None:
        memmap_dir = _get_output_data_location(tags)
        if not os.path.exists(memmap_dir):
            os.makedirs(memmap_dir)

        filename = str(cachestr) + fileending
        filename = os.path.join(memmap_dir, filename)
        return filename
    return None

def _load_data(cachestr, tags, fileending):
    """ loading data from scratch or from cache """
    filename = _get_cachefile(cachestr, tags, fileending)
    if filename is not None and os.path.exists(filename):
        return False
    return True

class GenomicArray(object):  # pylint: disable=too-many-instance-attributes
    """GenomicArray stores multi-dimensional genomic information.

    It acts as a dataset for holding genomic data. For instance,
    coverage along an entire genome composed of arbitrary length chromosomes
    as well as for multiple cell-types and conditions simultaneously.
    Inspired by the HTSeq analog, the array can hold the data in different
    storage modes, including ndarray, hdf5 or as sparse dataset.

    Parameters
    ----------
    stranded : bool
        Consider stranded profiles. Default: True.
    conditions : list(str) or None
        List of cell-type or condition labels associated with the corresponding
        array dimensions. Default: None means a one-dimensional array is produced.
    typecode : str
        Datatype. Default: 'd'.
    resolution : int or None
        Resolution for storing the genomic array. Only relevant for the use
        with Cover Datasets. Default: 1.
    padding_value : float
        Padding value. Default: 0.
    order : int
        Order of the alphabet size. Only relevant for Bioseq Datasets. Default: 1.
    collapser : None or callable
        Method to aggregate values along a given interval.
    """
    handle = OrderedDict()
    _condition = None
    _resolution = None
    _order = None
    region2index = None

    def __init__(self, stranded=True, conditions=None, typecode='d',
                 resolution=1, padding_value=0.,
                 order=1, store_whole_genome=True, collapser=None):
        self.stranded = stranded
        if conditions is None:
            conditions = ['sample']

        self.condition = conditions
        self.order = order
        self.padding_value = padding_value
        if not isinstance(order, int) or order < 1:
            raise Exception('order must be a positive integer')

        self.resolution = resolution
        self.typecode = typecode
        self._full_genome_stored = store_whole_genome
        self.collapser = collapser

    def _get_indices(self, interval, arraylen):
        """Given the original genomic coordinates,
           the array indices of the reference dataset (garray.handle)
           and the array indices of the view are returned.

        Parameters
        ----------
        interval : Interval
            Interval containing (chr, start, end)
        arraylen : int
            Length of the numpy target array.

        Returns
        -------
        tuple
            Tuple of indices corresponding to the slice in the arrays
            (ref_start, ref_end, array_start, array_end).
            ref_[start,end] indicate the slice in garray.handle
            while array_[start,end] indicates the slice in the
            target array / view.
        """
        chrom = interval.chrom
        start = self.get_iv_start(interval.start)
        end = self.get_iv_end(interval.end) - self.order + 1

        if start >= self.handle[chrom].shape[0]:
            return 0, 0, 0, 0
        else:
            array_start = 0
            ref_start = start

        if end > self.handle[chrom].shape[0]:
            array_end = arraylen - (end - self.handle[chrom].shape[0])
            ref_end = self.handle[chrom].shape[0]
        else:
            array_end = arraylen
            ref_end = end

        return ref_start, ref_end, array_start, array_end

    def __setitem__(self, index, value):
        interval = index[0]
        condition = index[1]
        if isinstance(condition, slice) and value.ndim != 3:
            raise ValueError('Expected 3D array with condition slice.')
        if isinstance(condition, slice):
            condition = slice(None, value.shape[-1], None)

        if self.stranded and value.shape[1] != 2:
            raise ValueError('If genomic array is in stranded mode, shape[-1] == 2 is expected')

        if not self.stranded and value.shape[1] != 1:
            value = value.sum(axis=1, keepdims=True)

        if isinstance(interval, Interval) and isinstance(condition, (int, slice)):
            start = self.get_iv_start(interval.start)
            end = self.get_iv_end(interval.end)

            length = end - start - self.order + 1
            # value should be a 2 dimensional array
            # it will be reshaped to a 2D array where the collapse operation is performed

            # along the second dimension.
            value = self._do_collapse(interval, value)

            try:
                self._setitem(interval, condition, length, value)

            except KeyError:
                # we end up here if the peak regions are not a subset of
                # the regions of interest. that might be the case if
                # peaks from the holdout proportion of the genome are tried
                # to be added.
                # unfortunately, it is also possible that store_whole_genome=False
                # and the peaks and regions of interest are just not synchronized
                # in which case nothing (or too few peaks) are added. in the latter
                # case an error would help actually, but I am not sure how to
                # check if the first or the second is the case here.
                pass
        else:
            raise IndexError("Cannot interpret interval and condition: {}".format((interval, condition)))

    def _setitem(self, interval, condition, length, value):
        if not self._full_genome_stored:
            idx = self.region2index[_iv_to_str(interval.chrom, interval.start,
                                               interval.end)]

            # correcting for the overshooting starts and ends is not necessary
            # for partially loaded data
            self.handle['data'][idx, :length, :, condition] = value

        else:
            ref_start, ref_end, array_start, \
                array_end = self._get_indices(interval, value.shape[0])
            self.handle[interval.chrom][ref_start:ref_end, :, condition] = \
                               value[array_start:array_end]

    def _do_collapse(self, interval, value):
        if self.collapser is not None:

            if self.resolution is None and value.shape[0] == 1 or \
                self.resolution is not None and \
                value.shape[0] == interval.length//self.resolution:
                # collapsing becomes obsolete, because the data has already
                # the expected shape (after collapsing)
                pass
            else:
                if self.resolution is None:
                    # collapse along the entire interval
                    value = value.reshape((1,) + value.shape)
                else:
                    # if the array shape[0] is a multipe of resolution,
                    # it can simply be reshaped. otherwise,
                    # it needs to be resized before.
                    if value.shape[0] % self.resolution > 0:
                        value = np.resize(value, (
                            int(np.ceil(value.shape[0]/float(self.resolution))*self.resolution),) +
                                          value.shape[1:])
                    # collapse in bins of size resolution
                    value = value.reshape((value.shape[0]//min(self.resolution,
                                                               value.shape[0]),
                                           min(self.resolution, value.shape[0]),) + \
                                          value.shape[1:])

                value = self.collapser(value)
        return value

    def __getitem__(self, index):
        # for now lets ignore everything except for chrom, start and end.
        if isinstance(index, Interval):
            interval = index
            chrom = interval.chrom
            start = self.get_iv_start(interval.start)
            end = self.get_iv_end(interval.end)

            # original length
            length = end-start - self.order + 1

            if not self._full_genome_stored:
                idx = self.region2index[_iv_to_str(chrom, interval.start, interval.end)]
                # correcting for the overshooting starts and ends is not necessary
                # for partially loaded data
                return self._reshape(self.handle['data'][idx],
                                     (length, 2 if self.stranded else 1,
                                      len(self.condition)))

            if chrom not in self.handle:
                return np.ones((length, 2 if self.stranded else 1,
                                len(self.condition)),
                               dtype=self.typecode) * self.padding_value

            if start >= 0 and end <= self.handle[chrom].shape[0]:
                end = end - self.order + 1
                # this is a short-cut, which does not require zero-padding
                return self._reshape(self.handle[chrom][start:end],
                                     (end-start, 2 if self.stranded else 1,
                                      len(self.condition)))

            # below is some functionality for zero-padding, in case the region
            # reaches out of the chromosome size

            if self.padding_value == 0.0:
                data = np.zeros((length, 2 if self.stranded else 1,
                                 len(self.condition)),
                                dtype=self.typecode)
            else:
                data = np.ones((length, 2 if self.stranded else 1,
                                len(self.condition)),
                               dtype=self.typecode) * self.padding_value

            ref_start, ref_end, array_start, array_end = self._get_indices(interval, data.shape[0])

            data[array_start:array_end, :, :] = self._reshape(self.handle[chrom][ref_start:ref_end],
                                                              (ref_end - ref_start,
                                                               2 if self.stranded else 1,
                                                               len(self.condition)))
            return data

        raise IndexError("Cannot interpret interval: {}".format(index))

    @property
    def condition(self):
        """condition"""
        return self._condition

    @condition.setter
    def condition(self, conditions):
        self._condition = conditions

    @property
    def resolution(self):
        """resolution"""
        return self._resolution

    @resolution.setter
    def resolution(self, value):
        if value is not None and value <= 0:
            raise ValueError('resolution>0 required')
        self._resolution = value

    def _reshape(self, data, shape):
        # shape not necessary here,
        # data should just fall through
        return data

    def interval_length(self, chrom):
        """Method returns the interval lengths."""
        # extract the length by the interval length
        # or by the array shape
        locus = _str_to_iv(chrom)
        if len(locus) > 1:
            return locus[2] - locus[1]

        return self.resolution

    def scale_by_region_length(self):
        """ This method scales the regions by the region length ."""
        for chrom in self.handle:
            if self._full_genome_stored:
                self.handle[chrom][:] /= self.interval_length(chrom)
            else:
                for rstr, idx in self.region2index.items():
                    self.handle[chrom][idx] /= self.interval_length(rstr)

    def weighted_mean(self):
        """ Base pair resolution mean weighted by interval length
        """

        # summing the signal
        sums = [self.sum(chrom) for chrom in self.handle]
        sums = np.asarray(sums).sum(axis=0)

        # weights are determined by interval and chromosome length
        weights = [np.prod(self.handle[chrom].shape[:-1]) \
                   for chrom in self.handle]
        weights = np.asarray(weights).sum()
        return sums / weights

    def shift(self, means):
        """Centering the signal by the weighted mean"""
        #means = self.weighted_mean()

        for chrom in self.handle:
            # adjust base pair resoltion mean to interval length
            self.handle[chrom][:] -= means

    def rescale(self, scale):
        """ Method to rescale the signal """
        for chrom in self.handle:
            self.handle[chrom][:] /= scale

    def sum(self, chrom=None):
        """Sum signal across chromosomes."""
        if chrom is not None:
            return self.handle[chrom][:]\
                .sum(axis=tuple(range(self.handle[chrom].ndim - 1)))

        return np.asarray([self.handle[chrom][:]\
            .sum(axis=tuple(range(self.handle[chrom].ndim - 1)))
                           for chrom in self.handle])

    def weighted_sd(self):
        """ Interval scaled standard deviation """

        # summing the squared signal signal
        sums = [np.square(self.handle[chrom][:, :, :]).sum(
            axis=tuple(range(self.handle[chrom].ndim - 1))) \
            for chrom in self.handle]
        sums = np.asarray(sums).sum(axis=0)

        # weights are determined by interval and chromosome length
        weights = [np.prod(self.handle[chrom].shape[:-1]) \
                   for chrom in self.handle]
        weights = np.asarray(weights).sum()
        return np.sqrt(sums / (weights - 1.))

    @property
    def order(self):
        """order"""
        return self._order

    @order.setter
    def order(self, order):
        if order <= 0:
            raise ValueError('order>0 required')
        self._order = order

    def get_iv_end(self, end):
        """obtain the chromosome length for a given resolution."""
        return _get_iv_length(end, self.resolution)

    def get_iv_start(self, start):
        """obtain the chromosome length for a given resolution."""
        if self.resolution is None:
            return 0
        return start // self.resolution

def init_with_padding_value(padding_value, shape, dtype):
    """ create array with given padding value. """
    if padding_value == 0.0:
        return np.zeros(shape, dtype)
    else:
        return np.ones(shape, dtype) * padding_value

class HDF5GenomicArray(GenomicArray):
    """HDF5GenomicArray stores multi-dimensional genomic information.

    Implements GenomicArray.

    Parameters
    ----------
    gsize : GenomicIndexer or callable
        GenomicIndexer containing the genome sizes or a callable that
        returns a GenomicIndexer to enable lazy loading.
    stranded : bool
        Consider stranded profiles. Default: True.
    conditions : list(str) or None
        List of cell-type or condition labels associated with the corresponding
        array dimensions. Default: None means a one-dimensional array is produced.
    typecode : str
        Datatype. Default: 'd'.
    datatags : list(str) or None
        Tags describing the dataset. This is used to store the cache file.
    resolution : int
        Resolution for storing the genomic array. Only relevant for the use
        with Cover Datasets. Default: 1.
    order : int
        Order of the alphabet size. Only relevant for Bioseq Datasets. Default: 1.
    store_whole_genome : boolean
        Whether to store the entire genome or only the regions of interest.
        Default: True
    padding_value : float
        Padding value. Default: 0.
    cache : str or None
        Hash string of the data and parameters to cache the dataset. If None,
        caching is deactivated. Default: None.
    overwrite : boolean
        Whether to overwrite the cache. Default: False
    loader : callable or None
        Function to be called for loading the genomic array.
    collapser : None or callable
        Method to aggregate values along a given interval.
    verbose : boolean
        Verbosity. Default: False
    """

    def __init__(self, gsize,  # pylint: disable=too-many-locals
                 stranded=True,
                 conditions=None,
                 typecode='d',
                 datatags=None,
                 resolution=1,
                 order=1,
                 padding_value=0.,
                 store_whole_genome=True,
                 cache=None,
                 overwrite=False, loader=None,
                 normalizer=None,
                 collapser=None,
                 verbose=False):
        super(HDF5GenomicArray, self).__init__(stranded, conditions, typecode,
                                               resolution,
                                               order=order,
                                               padding_value=padding_value,
                                               store_whole_genome=store_whole_genome,
                                               collapser=collapser)

        if cache is None:
            raise ValueError('cache=True required for HDF format')

        gsize_ = None

        if not store_whole_genome:
            gsize_ = gsize() if callable(gsize) else gsize
            self.region2index = {_iv_to_str(region.chrom,
                                            region.start,
                                            region.end): i \
                                                for i, region in enumerate(gsize_)}

        cachefile = _get_cachefile(cache, datatags, '.h5')
        load_from_file = _load_data(cache, datatags, '.h5')

        if load_from_file:
            if gsize_ is None:
                gsize_ = gsize() if callable(gsize) else gsize

            h5file = h5py.File(cachefile, 'w')

            if store_whole_genome:
                for region in gsize_:
                    shape = (_get_iv_length(region.length - self.order + 1, self.resolution),
                             2 if stranded else 1, len(self.condition))
                    h5file.create_dataset(str(region.chrom), shape,
                                          dtype=self.typecode,
                                          data=init_with_padding_value(padding_value,
                                                                       shape,
                                                                       self.typecode))
                    self.handle = h5file
            else:
                shape = (len(gsize_),
                         _get_iv_length(gsize_.binsize + 2*gsize_.flank - self.order + 1,
                                        self.resolution),
                         2 if stranded else 1, len(self.condition))
                h5file.create_dataset('data', shape,
                                      dtype=self.typecode,
                                      data=init_with_padding_value(padding_value,
                                                                   shape,
                                                                   self.typecode))
                self.handle = h5file
            # invoke the loader
            if loader:
                loader(self)

            for norm in normalizer or []:
                get_normalizer(norm)(self)
            h5file.close()
        if verbose: print('reload {}'.format(cachefile))
        h5file = h5py.File(cachefile, 'a', driver='stdio')

        self.handle = h5file


class NPGenomicArray(GenomicArray):
    """NPGenomicArray stores multi-dimensional genomic information.

    Implements GenomicArray.
    Parameters
    ----------
    gsize : GenomicIndexer or callable
        GenomicIndexer containing the genome sizes or a callable that
        returns a GenomicIndexer to enable lazy loading.
    stranded : bool
        Consider stranded profiles. Default: True.
    conditions : list(str) or None
        List of cell-type or condition labels associated with the corresponding
        array dimensions. Default: None means a one-dimensional array is produced.
    typecode : str
        Datatype. Default: 'd'.
    datatags : list(str) or None
        Tags describing the dataset. This is used to store the cache file.
    resolution : int
        Resolution for storing the genomic array. Only relevant for the use
        with Cover Datasets. Default: 1.
    order : int
        Order of the alphabet size. Only relevant for Bioseq Datasets. Default: 1.
    store_whole_genome : boolean
        Whether to store the entire genome or only the regions of interest.
        Default: True
    padding_value : float
        Padding value. Default: 0.
    cache : str or None
        Hash string of the data and parameters to cache the dataset. If None,
        caching is deactivated. Default: None.
    overwrite : boolean
        Whether to overwrite the cache. Default: False
    loader : callable or None
        Function to be called for loading the genomic array.
    normalizer : callable or None
        Normalization to be applied. This argumenet can be None,
        if no normalization is applied, or a callable that takes
        a garray and returns a normalized garray.
        Default: None.
    collapser : None or callable
        Method to aggregate values along a given interval.
    verbose : boolean
        Verbosity. Default: False
    """

    def __init__(self, gsize,  # pylint: disable=too-many-locals
                 stranded=True,
                 conditions=None,
                 typecode='d',
                 datatags=None,
                 resolution=1,
                 order=1,
                 padding_value=0.0,
                 store_whole_genome=True,
                 cache=None,
                 overwrite=False, loader=None,
                 normalizer=None, collapser=None,
                 verbose=False):

        super(NPGenomicArray, self).__init__(stranded, conditions, typecode,
                                             resolution,
                                             order=order,
                                             padding_value=padding_value,
                                             store_whole_genome=store_whole_genome,
                                             collapser=collapser)

        gsize_ = None

        if not store_whole_genome:

            gsize_ = gsize() if callable(gsize) else gsize

            self.region2index = {_iv_to_str(region.chrom,
                                            region.start,
                                            region.end): i \
                                                for i, region in enumerate(gsize_)}

        cachefile = _get_cachefile(cache, datatags, '.npz')
        load_from_file = _load_data(cache, datatags, '.npz')

        if load_from_file:
            if gsize_ is None:
                gsize_ = gsize() if callable(gsize) else gsize

            if store_whole_genome:
                data = {str(region.chrom): init_with_padding_value(
                    padding_value,
                    shape=(_get_iv_length(region.length - self.order + 1,
                                          self.resolution),
                           2 if stranded else 1,
                           len(self.condition)),
                    dtype=self.typecode) for region in gsize_}
                names = [str(region.chrom) for region in gsize_]
                self.handle = data
            else:
                data = {'data': init_with_padding_value(
                    padding_value,
                    shape=(len(gsize_),
                           _get_iv_length(gsize_.binsize + 2*gsize_.flank - self.order + 1,
                                          self.resolution) if self.resolution is not None else 1,
                           2 if stranded else 1,
                           len(self.condition)),
                    dtype=self.typecode)}
                names = ['data']
                self.handle = data

            # invoke the loader
            if loader:
                loader(self)


            if cachefile is not None:
                np.savez(cachefile, **data)


        if cachefile is not None:
            if verbose: print('reload {}'.format(cachefile))
            data = np.load(cachefile)
            names = [x for x in data]

        # here we get either the freshly loaded data or the reloaded
        # data from np.load.
        self.handle = {key: data[key] for key in names}

        for norm in normalizer or []:
            get_normalizer(norm)(self)


class SparseGenomicArray(GenomicArray):
    """SparseGenomicArray stores multi-dimensional genomic information.

    Implements GenomicArray.

    Parameters
    ----------
    gsize : GenomicIndexer or callable
        GenomicIndexer containing the genome sizes or a callable that
        returns a GenomicIndexer to enable lazy loading.
    stranded : bool
        Consider stranded profiles. Default: True.
    conditions : list(str) or None
        List of cell-type or condition labels associated with the corresponding
        array dimensions. Default: None means a one-dimensional array is produced.
    typecode : str
        Datatype. Default: 'd'.
    datatags : list(str) or None
        Tags describing the dataset. This is used to store the cache file.
    resolution : int
        Resolution for storing the genomic array. Only relevant for the use
        with Cover Datasets. Default: 1.
    order : int
        Order of the alphabet size. Only relevant for Bioseq Datasets. Default: 1.
    store_whole_genome : boolean
        Whether to store the entire genome or only the regions of interest.
        Default: True
    padding_value : float
        Padding value. Default: 0.
    cache : str or None
        Hash string of the data and parameters to cache the dataset. If None,
        caching is deactivated. Default: None.
    overwrite : boolean
        Whether to overwrite the cache. Default: False
    loader : callable or None
        Function to be called for loading the genomic array.
    normalizer : callable or None
        Normalization to be applied. This argumenet can be None,
        if no normalization is applied, or a callable that takes
        a garray and returns a normalized garray.
        Default: None.
    collapser : None or callable
        Method to aggregate values along a given interval.
    verbose : boolean
        Verbosity. Default: False
    """

    def __init__(self, gsize,  # pylint: disable=too-many-locals
                 stranded=True,
                 conditions=None,
                 typecode='d',
                 datatags=None,
                 resolution=1,
                 order=1,
                 store_whole_genome=True,
                 cache=None,
                 padding_value=0.0,
                 overwrite=False,
                 loader=None,
                 collapser=None,
                 verbose=False):
        super(SparseGenomicArray, self).__init__(stranded, conditions,
                                                 typecode,
                                                 resolution,
                                                 order=order,
                                                 padding_value=padding_value,
                                                 store_whole_genome=store_whole_genome,
                                                 collapser=collapser)

        cachefile = _get_cachefile(cache, datatags, '.npz')
        load_from_file = _load_data(cache, datatags, '.npz')

        gsize_ = None

        if not store_whole_genome:

            gsize_ = gsize() if callable(gsize) else gsize

            self.region2index = {_iv_to_str(region.chrom,
                                            region.start,
                                            region.end): i \
                                                for i, region in enumerate(gsize_)}

        if load_from_file:
            if gsize_ is None:
                gsize_ = gsize() if callable(gsize) else gsize

            if store_whole_genome:
                data = {str(region.chrom): sparse.dok_matrix(
                    (_get_iv_length(region.length - self.order + 1,
                                    resolution),
                     (2 if stranded else 1) * len(self.condition)),
                    dtype=self.typecode)
                        for region in gsize_}
            else:
                data = {'data': sparse.dok_matrix(
                    (len(gsize_),
                     (_get_iv_length(gsize_.binsize + 2*gsize_.flank - self.order + 1,
                                     self.resolution) if self.resolution is not None else 1) *
                     (2 if stranded else 1) * len(self.condition)),
                    dtype=self.typecode)}
            self.handle = data

            # invoke the loader
            if loader:
                loader(self)

            data = self.handle

            data = {chrom: data[chrom].tocoo() for chrom in data}

            storage = {chrom: np.column_stack([data[chrom].data,
                                               data[chrom].row,
                                               data[chrom].col]) \
                                               for chrom in data}
            for region in gsize_:
                if store_whole_genome:
                    storage[region.chrom + '__length__'] = region.length

            names = [name for name in storage]

            if cachefile is not None:
                np.savez(cachefile, **storage)

        if cachefile is not None:
            if verbose: print('reload {}'.format(cachefile))
            storage = np.load(cachefile)

        names = [name for name in storage if '__length__' not in name]

        if store_whole_genome:
            self.handle = {name: sparse.coo_matrix(
                (storage[name][:, 0],
                 (storage[name][:, 1].astype('int'),
                  storage[name][:, 2].astype('int'))),
                shape=(_get_iv_length(storage[str(name)+'__length__'], resolution),
                       (2 if stranded else 1) * len(self.condition))).tocsr()
                           for name in names}
        else:
            # gsize_ is always available for store_whole_genome=False
            self.handle = {name: sparse.coo_matrix(
                (storage[name][:, 0],
                 (storage[name][:, 1].astype('int'),
                  storage[name][:, 2].astype('int'))),
                shape=(len(gsize_),
                       (_get_iv_length(gsize_.binsize + 2*gsize_.flank, resolution)
                        if self.resolution is not None else 1) *
                       (2 if stranded else 1) * len(self.condition))).tocsr()
                           for name in names}

    def _reshape(self, data, shape):
        # what to do with zero padding
        data = data.toarray()

        if self._full_genome_stored:
            return data.reshape(data.shape[0], data.shape[1]//(shape[-1]), shape[-1])
        else:
            return data.reshape(data.shape[1]//(shape[-2]*shape[-1]), shape[-2], shape[-1])

    def _setitem(self, interval, condition, length, value):
        if not self._full_genome_stored:
            regidx = self.region2index[_iv_to_str(interval.chrom, interval.start, interval.end)]
            nconditions = len(self.condition)
            ncondstrand = len(self.condition) * value.shape[-1]
            #end = end - self.order + 1
            idxs = np.where(value > 0)
            for idx in zip(*idxs):
                basepos = idx[0] * ncondstrand
                strand = idx[1] * nconditions
                cond = condition if isinstance(condition, int) else idx[2]
                self.handle['data'][regidx,
                                    basepos + strand + cond] = value[idx]
        else:
            ref_start, ref_end, array_start, _ = self._get_indices(interval, value.shape[0])
            idxs = np.where(value > 0)
            iarray = np.arange(ref_start, ref_end)
            for idx in zip(*idxs):
                cond = condition if isinstance(condition, int) else idx[2]
                self.handle[interval.chrom][iarray[idx[0]],
                                            idx[1] * len(self.condition)
                                            + cond] = value[idx[0] + array_start][idx[1:]]

class PercentileTrimming(object):
    """Percentile trimming normalization.

    This class performs percentile trimming of a GenomicArray to aleviate
    the effect of outliers.
    All values that exceed the value associated with the given percentile
    are set to be equal to the percentile.

    Parameters
    ----------
    percentile : float
        Percentile at which to perform chromosome-level trimming.
    """
    def __init__(self, percentile):
        self.percentile = percentile

    def __call__(self, garray):

        quants = np.percentile(np.concatenate(np.asarray([garray.handle[chrom] for \
                                      chrom in garray.handle]), axis=0),
                               self.percentile, axis=(0, 1))

        for icond, quant in enumerate(quants):
            for chrom in garray.handle:
                arr = garray.handle[chrom][:, :, icond]
                arr[arr > quant] = quant
                garray.handle[chrom][:, :, icond] = arr
        return garray

    def __str__(self):  # pragma: no cover
        return 'PercentileTrimming({})'.format(self.percentile)

    def __repr__(self):  # pragma: no cover
        return str(self)


class RegionLengthNormalization(object):
    """ Normalization for variable-region length.

    This class performs region length normalization of a GenomicArray.
    This is relevant when genomic features are of variable size, e.g.
    enhancer regions of different width or when using variable length genes.

    Parameters
    ----------
    regionmask : str or GenomicIndexer, None
        A bed file or a genomic indexer that contains the masking region
        that is considered for the signal. For instance, when normalizing
        gene expression to TPM, the mask contains exons. Otherwise, the
        TPM would normalize for the full length gene annotation.
        If None, no mask is included.
    """
    def __init__(self, regionmask=None):
        self.regionmask = regionmask

    def __call__(self, garray):
        # length scaling

        garray.scale_by_region_length()

        return garray

    def __str__(self):  # pragma: no cover
        return 'RegionLengthNormalization({})'.format(self.regionmask)

    def __repr__(self):  # pragma: no cover
        return str(self)


class ZScore(object):
    """ZScore normalization.

    This class performs ZScore normalization of a GenomicArray.
    It automatically adjusts for variable interval lenths.

    Parameters
    ----------
    means : float or None
        Provided means will be applied for zero-centering.
        If None, the means will be determined
        from the GenomicArray and then applied.
        Default: None.
    stds : float or None
        Provided standard deviations will be applied for scaling.
        If None, the stds will be determined
        from the GenomicArray and then applied.
        Default: None.
    """
    def __init__(self, mean=None, std=None):
        self.mean = mean
        self.std = std

    def __call__(self, garray):

        # determine the mean signal per condition
        if self.mean is None:
            self.mean = garray.weighted_mean()

        # centering to zero-mean
        garray.shift(self.mean)

        # determines standard deviation per contition
        if self.std is None:
            self.std = garray.weighted_sd()

        # rescale by standard deviation
        garray.rescale(self.std)

        return garray

    def __str__(self):  # pragma: no cover
        return 'ZScore({},{})'.format(self.mean, self.std)

    def __repr__(self):  # pragma: no cover
        return str(self)


class LogTransform(object):
    """Log transformation of intput signal.

    This class performs log-transformation
    of a GenomicArray using log(x + 1.) to avoid NAN's from zeros.

    """
    def __init__(self):
        pass

    def __call__(self, garray):

        for chrom in garray.handle:
            garray.handle[chrom][:] = np.log(garray.handle[chrom][:] + 1.)
        return garray

    def __str__(self):  # pragma: no cover
        return 'Log'

    def __repr__(self):  # pragma: no cover
        return str(self)


class ZScoreLog(object):
    """ZScore normalization after log transformation.

    This class performs ZScore normalization after log-transformation
    of a GenomicArray using log(x + 1.) to avoid NAN's from zeros.
    It automatically adjusts for variable interval lenths.

    Parameters
    ----------
    means : float or None
        Provided means will be applied for zero-centering.
        If None, the means will be determined
        from the GenomicArray and then applied.
        Default: None.
    stds : float or None
        Provided standard deviations will be applied for scaling.
        If None, the stds will be determined
        from the GenomicArray and then applied.
        Default: None.
    """
    def __init__(self, mean=None, std=None):
        self.logtr = LogTransform()
        self.zscore = ZScore(mean, std)

    def __call__(self, garray):
        return self.zscore(self.logtr(garray))

    def __str__(self):  # pragma: no cover
        return str(self.zscore) + str(self.logtr)

    def __repr__(self):  # pragma: no cover
        return str(self)


def normalize_garray_tpm(garray):
    """This function performs TPM normalization
    for a given GenomicArray.

    """

    # rescale by region lengths in bp
    garray = RegionLengthNormalization()(garray)

    # recale to kb
    garray.rescale(1e-3)

    # compute scaling factor
    scale = garray.sum() # per chromsome sum
    scale = scale.sum(axis=0) # sum across chroms
    scale /= 1e6 # rescale by million

    # divide by scaling factor
    garray.rescale(scale)

    return garray


def get_normalizer(normalizer):
    """ maps built-in normalizers by name and
    returns the respective function """

    if callable(normalizer):
        return normalizer
    elif normalizer == 'zscore':
        return ZScore()
    elif normalizer == 'zscorelog':
        return ZScoreLog()
    elif normalizer == 'binsizenorm':
        return RegionLengthNormalization()
    elif normalizer == 'perctrim':
        return PercentileTrimming(99)
    elif normalizer == 'tpm':
        return normalize_garray_tpm

    raise ValueError('unknown normalizer: {}'.format(normalizer))


def create_genomic_array(chroms, stranded=True, conditions=None, typecode='float32',
                         storage='hdf5', resolution=1,
                         order=1,
                         padding_value=0.0,
                         store_whole_genome=True,
                         datatags=None, cache=None, overwrite=False,
                         loader=None,
                         normalizer=None, collapser=None,
                         verbose=False):
    """Factory function for creating a GenomicArray.

    This function creates a genomic array for a given storage mode.

    Parameters
    ----------
    chroms : dict
        Dictionary with chromosome names as keys and chromosome lengths
        as values.
    stranded : bool
        Consider stranded profiles. Default: True.
    conditions : list(str) or None
        List of cell-type or condition labels associated with the corresponding
        array dimensions. Default: None means a one-dimensional array is produced.
    typecode : str
        Datatype. Default: 'float32'.
    storage : str
        Storage type can be 'ndarray', 'hdf5' or 'sparse'.
        Numpy loads the entire dataset into the memory. HDF5 keeps
        the data on disk and loads the mini-batches from disk.
        Sparse maintains sparse matrix representation of the dataset
        in the memory.
        Usage of numpy will require high memory consumption, but allows fast
        slicing operations on the dataset. HDF5 requires low memory consumption,
        but fetching the data from disk might be time consuming.
        sparse will be a good compromise if the data is indeed sparse. In this
        case, memory consumption will be low while slicing will still be fast.
    datatags : list(str) or None
        Tags describing the dataset. This is used to store the cache file.
    resolution : int
        Resolution for storing the genomic array. Only relevant for the use
        with Cover Datasets. Default: 1.
    order : int
        Order of the alphabet size. Only relevant for Bioseq Datasets. Default: 1.
    padding_value : float
        Padding value. Default: 0.0
    store_whole_genome : boolean
        Whether to store the entire genome or only the regions of interest.
        Default: True
    cache : str or None
        Hash string of the data and parameters to cache the dataset. If None,
        caching is deactivated. Default: None.
    overwrite : boolean
        Whether to overwrite the cache. Default: False
    loader : callable or None
        Function to be called for loading the genomic array.
    normalizer : callable or None
        Normalization to be applied. This argumenet can be None,
        if no normalization is applied, or a callable that takes
        a garray and returns a normalized garray.
        Default: None.
    collapser : str, callable or None
        Collapse method defines how the signal is aggregated for resolution>1 or resolution=None.
        For example, by summing the signal over a given interval.
    verbose : boolean
        Verbosity. Default: False
    """

    # check if collapser available
    if (resolution is None or resolution > 1) and collapser is None:
        raise ValueError('Requiring collapser due to resolution=None or resolution>1')

    # force store_whole_genome=False if resolution=None
    if resolution is None and store_whole_genome:
        print('store_whole_genome=True ignored, because it is not compatible'
              'with resolution=None. store_whole_genome=False is used instead.')
        store_whole_genome = False

    if storage == 'hdf5':
        return HDF5GenomicArray(chroms, stranded=stranded,
                                conditions=conditions,
                                typecode=typecode,
                                datatags=datatags,
                                resolution=resolution,
                                order=order,
                                store_whole_genome=store_whole_genome,
                                cache=cache,
                                padding_value=padding_value,
                                overwrite=overwrite,
                                loader=loader,
                                normalizer=normalizer,
                                collapser=get_collapser(collapser),
                                verbose=verbose)
    elif storage == 'ndarray':
        return NPGenomicArray(chroms, stranded=stranded,
                              conditions=conditions,
                              typecode=typecode,
                              datatags=datatags,
                              resolution=resolution,
                              order=order,
                              store_whole_genome=store_whole_genome,
                              cache=cache,
                              padding_value=padding_value,
                              overwrite=overwrite,
                              loader=loader,
                              normalizer=normalizer,
                              collapser=get_collapser(collapser),
                              verbose=verbose)
    elif storage == 'sparse':
        return SparseGenomicArray(chroms, stranded=stranded,
                                  conditions=conditions,
                                  typecode=typecode,
                                  datatags=datatags,
                                  resolution=resolution,
                                  order=order,
                                  store_whole_genome=store_whole_genome,
                                  cache=cache,
                                  padding_value=padding_value,
                                  overwrite=overwrite,
                                  loader=loader,
                                  collapser=get_collapser(collapser),
                                  verbose=verbose)

    raise Exception("Storage type must be 'hdf5', 'ndarray' or 'sparse'")




import os

from pkg_resources import resource_filename

from janggu.data import GenomicIndexer
from janggu.janggutrim import trim_bed


def test_create_from_array(tmpdir):
    inbed = resource_filename('janggu', 'resources/bed_test.bed')

    outbed = os.path.join(tmpdir.strpath, 'out.bed')
    trim_bed(inbed, outbed, 5)

    # original file
    gindexer = GenomicIndexer.create_from_file(inbed, None, None)
    reg = gindexer[0]
    assert (reg.start % 5) == 0
    assert (reg.end % 5) > 0

    # trimmed file
    gindexer = GenomicIndexer.create_from_file(outbed, None, None)
    gindexer = GenomicIndexer.create_from_file(outbed, None, None)
    reg = gindexer[0]
    assert (reg.start % 5) == 0
    assert (reg.end % 5) == 0



"""Genomic track visualization utilities."""
import warnings
from itertools import product

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from janggu.utils import NMAP
from janggu.utils import PMAP
from janggu.utils import _to_list


def plotGenomeTrack(tracks, chrom, start, end, figsize=(10, 5), plottypes=None):

    """plotGenomeTrack shows plots of a specific interval from cover objects data.

    It takes one or more cover objects as well as a genomic interval consisting
    of chromosome name, start and end and creates
    a genome browser-like plot.

    Parameters
    ----------
    tracks : janggu.data.Cover, list(Cover), janggu.data.Track or list(Track)
        One or more track objects.
    chrom : str
        chromosome name.
    start : int
        The start of the required interval.
    end : int
        The end of the required interval.
    figsize : tuple(int, int)
        Figure size passed on to matplotlib.
    plottype : None or list(str)
        Plot type indicates whether to plot coverage tracks as line plots,
        heatmap, or seqplot using 'line' or 'heatmap', respectively.
        By default, all coverage objects are depicted as line plots if plottype=None.
        Otherwise, a list of types must be supplied containing the plot types for each
        coverage object explicitly. For example, ['line', 'heatmap', 'seqplot'].
        While, 'line' and 'heatmap' can be used for any type of coverage data,
        'seqplot' is reserved to plot sequence influence on the output. It is
        intended to be used in conjunction with 'input_attribution' method which
        determines the importance of paricular sequence letters for the output prediction.

    Returns
    -------
    matplotlib Figure
        A matplotlib figure illustrating the genome browser-view of the coverage
        objects for the given interval.
        To depict and save the figure the native matplotlib functions show()
        and savefig() can be used.
    """

    tracks = _to_list(tracks)

    for track in tracks:
        if not isinstance(track, Track):
            warnings.warn('Convert the Dataset object to proper Track objects.'
                          ' In the future, only Track objects will be supported.',
                          FutureWarning)
            if plottypes is None:
                plottypes = ['line'] * len(tracks)

            assert len(plottypes) == len(tracks), \
                "The number of cover objects must be the same as the number of plottyes."
            break

    def _convert_to_track(cover, plottype):
        if plottype == 'heatmap':
            track = HeatTrack(cover)
        elif plottype == 'seqplot':
            track = SeqTrack(cover)
        else:
            track = LineTrack(cover)
        return track

    tracks_ = []
    for itrack, track in enumerate(tracks):
        if isinstance(track, Track):
            tracks_.append(track)
        else:
            warnings.warn('Convert the Dataset object to proper Track objects.'
                          ' In the future, only Track objects will be supported.',
                          FutureWarning)
            tracks_.append(_convert_to_track(track, plottypes[itrack]))

    tracks = tracks_
    headertrack = 2
    trackheights = 0
    for track in tracks:
        trackheights += track.height
    spacer = len(tracks) - 1

    grid = plt.GridSpec(headertrack + trackheights + spacer,
                        10, wspace=0.4, hspace=0.3)
    fig = plt.figure(figsize=figsize)

    # title and reference track
    title = fig.add_subplot(grid[0, 1:])

    title.set_title(chrom)
    plt.xlim([0, end - start])
    title.spines['right'].set_visible(False)
    title.spines['top'].set_visible(False)
    title.spines['left'].set_visible(False)
    plt.xticks([0, end-start], [start, end])
    plt.yticks(())

    y_offset = 1
    for track in tracks:
        y_offset += 1

        track.add_side_bar(fig, grid, y_offset)
        track.plot(fig, grid, y_offset, chrom, start, end)
        y_offset += track.height

    return (fig)


class Track(object):
    """General track

    Parameters
    ----------

    data : Cover object
        Coverage object
    height : int
        Track height.
    """
    def __init__(self, data, height):
        self.height = height
        self.data = data

    @property
    def name(self):
        """Track name"""
        return self.data.name

    def add_side_bar(self, fig, grid, offset):
        """Side-bar"""
        # side bar indicator for current cover
        ax = fig.add_subplot(grid[(offset): (offset + self.height), 0])

        ax.set_xticks(())
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.spines['bottom'].set_visible(False)
        ax.set_yticks([0.5])
        ax.set_yticklabels([self.name])

    def get_track_axis(self, fig, grid, offset, height):
        """Returns axis object"""
        return fig.add_subplot(grid[offset:(offset + height), 1:])

    def get_data(self, chrom, start, end):
        """Returns data to plot."""
        return self.data[chrom, start, end][0, :, :, :]


class LineTrack(Track):
    """Line track

    Visualizes genomic data as line plot.

    Parameters
    ----------

    data : Cover object
        Coverage object
    height : int
        Track height. Default=3
    linestyle : str
        Linestyle for plot
    marker : str
        Marker code for plot
    color : str
        Color code for plot
    linewidth : float
        Line width.
    """
    def __init__(self, data, height=3, linestyle='-', marker='o', color='b',
                 linewidth=2):
        super(LineTrack, self).__init__(data, height)
        self.height = height * len(data.conditions)
        self.linestyle = linestyle
        self.linewidth = linewidth
        self.marker = marker
        self.color = color

    def plot(self, fig, grid, offset, chrom, start, end):
        """Plot line track."""
        coverage = self.get_data(chrom, start, end)
        offset_ = offset
        trackheight = self.height//len(self.data.conditions)

        def _get_xy(cov):
            xvalue = np.where(np.isfinite(cov))[0]
            yvalue = cov[xvalue]
            return xvalue, yvalue

        for i, condition in enumerate(self.data.conditions):
            ax = self.get_track_axis(fig, grid, offset_, trackheight)
            offset_ += trackheight
            if coverage.shape[1] == 2:
                #both strands are covered separately
                xvalue, yvalue = _get_xy(coverage[:, 0, i])
                ax.plot(xvalue, yvalue,
                        linewidth=self.linewidth,
                        linestyle=self.linestyle,
                        color=self.color, label="+", marker='+')
                xvalue, yvalue = _get_xy(coverage[:, 1, i])
                ax.plot(xvalue, yvalue,
                        linewidth=self.linewidth,
                        linestyle=self.linestyle,
                        color=self.color, label="-", marker=1)
                ax.legend()
            else:
                xvalue, yvalue = _get_xy(coverage[:, 0, i])
                ax.plot(xvalue, yvalue, linewidth=self.linewidth,
                        color=self.color,
                        linestyle=self.linestyle,
                        marker=self.marker)
            ax.set_yticks(())
            ax.set_xticks(())
            ax.set_xlim([0, end-start])
            if len(self.data.conditions) > 1:
                ax.set_ylabel(condition, labelpad=12)
            ax.spines['right'].set_visible(False)
            ax.spines['top'].set_visible(False)


class SeqTrack(Track):
    """Sequence Track

    Visualizes sequence importance.

    Parameters
    ----------

    data : Cover object
        Coverage object
    height : int
        Track height. Default=3
    """
    def __init__(self, data, height=3):
        super(SeqTrack, self).__init__(data, height)

    def plot(self, fig, grid, offset, chrom, start, end):
        """Plot sequence track"""

        if len(self.data.conditions) % len(NMAP) == 0:
            alphabetsize = len(NMAP)
            MAP = NMAP
        elif len(self.data.conditions) % len(PMAP) == 0:  # pragma: no cover
            alphabetsize = len(PMAP)
            MAP = PMAP
        else:  # pragma: no cover
            raise ValueError(
                "Coverage tracks seems not represent biological sequences. "
                "The last dimension must be divisible by the alphabetsize.")

        for cond in self.data.conditions:
            if cond[0] not in MAP:
                raise ValueError(
                    "Coverage tracks seems not represent biological sequences. "
                    "Condition names must represent the alphabet letters.")

        coverage = self.get_data(chrom, start, end)
        # project higher-order sequence structure onto original sequence.
        coverage = coverage.reshape(coverage.shape[0], -1)
        coverage = coverage.reshape(coverage.shape[:-1] +
                                    (alphabetsize,
                                     int(coverage.shape[-1]/alphabetsize)))
        coverage = coverage.sum(-1)

        ax = self.get_track_axis(fig, grid, offset, self.height)
        x = np.arange(coverage.shape[0])
        y_figure_offset = np.zeros(coverage.shape[0])
        handles = []
        for letter in MAP:
            handles.append(ax.bar(x, coverage[:, MAP[letter]],
                                  bottom=y_figure_offset,
                                  color=sns.color_palette("hls",
                                                          len(MAP))[MAP[letter]],
                                  label=letter))
            y_figure_offset += coverage[:, MAP[letter]]
        ax.legend(handles=handles)
        ax.set_yticklabels(())
        ax.set_yticks(())
        ax.set_xticks(())
        ax.set_xlim([0, end-start])


class HeatTrack(Track):
    """Heatmap Track

    Visualizes genomic data as heatmap.

    Parameters
    ----------

    data : Cover object
        Coverage object
    height : int
        Track height. Default=3
    """
    def __init__(self, data, height=3):
        super(HeatTrack, self).__init__(data, height)

    def plot(self, fig, grid, offset, chrom, start, end):
        """Plot heatmap track."""
        ax = self.get_track_axis(fig, grid, offset, self.height)
        coverage = self.get_data(chrom, start, end)

        im = ax.pcolor(coverage.reshape(coverage.shape[0], -1).T)

        if coverage.shape[-2] == 2:
            ticks = [':'.join([x, y]) for y, x \
                     in product(['+', '-'], self.data.conditions)]
        else:
            ticks = self.data.conditions

        ax.set_yticklabels(ticks)
        ax.set_xticks(())
        ax.set_yticks(np.arange(0, len(ticks) + 1, 1.0))
        ax.set_xlim([0, end-start])
